# coding: utf-8
"""create or manipulate already existing labeling jobs"""

# docs
from __future__ import annotations

import time
from typing import TYPE_CHECKING, Callable, Dict, List, NamedTuple, Optional, Union

from tqdm import tqdm

if TYPE_CHECKING:
    from pandas.core.frame import DataFrame

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.api.entity_annotation.figure_api import FigureInfo
from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    ModuleWithStatus,
    RemoveableBulkModuleApi,
    WaitingTimeExceeded,
)
from supervisely.collection.str_enum import StrEnum
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.project.project_meta import ProjectMeta


class LabelingJobInfo(NamedTuple):
    id: int
    name: str
    readme: str
    description: str
    team_id: int
    workspace_id: int
    workspace_name: str
    project_id: int
    project_name: str
    dataset_id: int
    dataset_name: str
    created_by_id: int
    created_by_login: str
    assigned_to_id: int
    assigned_to_login: str
    reviewer_id: int
    reviewer_login: str
    created_at: str
    started_at: str
    finished_at: str
    status: str
    disabled: bool
    images_count: int
    finished_images_count: int
    rejected_images_count: int
    accepted_images_count: int
    progress_images_count: int
    classes_to_label: list
    tags_to_label: list
    images_range: tuple
    objects_limit_per_image: int
    tags_limit_per_image: int
    filter_images_by_tags: list
    include_images_with_tags: list
    exclude_images_with_tags: list
    entities: list


class LabelingJobApi(RemoveableBulkModuleApi, ModuleWithStatus):
    """
    API for working with Labeling Jobs. :class:`LabelingJobApi<LabelingJobApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        jobs = api.labeling_job.get_list(9) # api usage example
    """

    class Status(StrEnum):
        """Labeling Job status."""

        PENDING = "pending"
        """"""
        IN_PROGRESS = "in_progress"
        """"""
        ON_REVIEW = "on_review"
        """"""
        COMPLETED = "completed"
        """"""
        STOPPED = "stopped"
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple LabelingJobInfo information about Labeling Job.

        :Example:

         .. code-block:: python

             LabelingJobInfo(id=2,
                             name='Annotation Job (#1) (#1) (dataset_01)',
                             readme='',
                             description='',
                             team_id=4,
                             workspace_id=8,
                             workspace_name='First Workspace',
                             project_id=58,
                             project_name='tutorial_project',
                             dataset_id=54,
                             dataset_name='dataset_01',
                             created_by_id=4,
                             created_by_login='anna',
                             assigned_to_id=4,
                             assigned_to_login='anna',
                             reviewer_id=4,
                             reviewer_login='anna',
                             created_at='2020-04-08T15:10:12.618Z',
                             started_at='2020-04-08T15:10:19.833Z',
                             finished_at='2020-04-08T15:13:39.788Z',
                             status='completed',
                             disabled=False,
                             images_count=3,
                             finished_images_count=0,
                             rejected_images_count=1,
                             accepted_images_count=2,
                             progress_images_count=2,
                             classes_to_label=[],
                             tags_to_label=[],
                             images_range=(1, 5),
                             objects_limit_per_image=None,
                             tags_limit_per_image=None,
                             filter_images_by_tags=[],
                             include_images_with_tags=[],
                             exclude_images_with_tags=[],
                             entities=None)
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.README,
            ApiField.DESCRIPTION,
            ApiField.TEAM_ID,
            ApiField.WORKSPACE_ID,
            ApiField.WORKSPACE_NAME,
            ApiField.PROJECT_ID,
            ApiField.PROJECT_NAME,
            ApiField.DATASET_ID,
            ApiField.DATASET_NAME,
            ApiField.CREATED_BY_ID,
            ApiField.CREATED_BY_LOGIN,
            ApiField.ASSIGNED_TO_ID,
            ApiField.ASSIGNED_TO_LOGIN,
            ApiField.REVIEWER_ID,
            ApiField.REVIEWER_LOGIN,
            ApiField.CREATED_AT,
            ApiField.STARTED_AT,
            ApiField.FINISHED_AT,
            ApiField.STATUS,
            ApiField.DISABLED,
            ApiField.IMAGES_COUNT,
            ApiField.FINISHED_IMAGES_COUNT,
            ApiField.REJECTED_IMAGES_COUNT,
            ApiField.ACCEPTED_IMAGES_COUNT,
            ApiField.PROGRESS_IMAGES_COUNT,
            ApiField.CLASSES_TO_LABEL,
            ApiField.TAGS_TO_LABEL,
            ApiField.IMAGES_RANGE,
            ApiField.OBJECTS_LIMIT_PER_IMAGE,
            ApiField.TAGS_LIMIT_PER_IMAGE,
            ApiField.FILTER_IMAGES_BY_TAGS,
            ApiField.INCLUDE_IMAGES_WITH_TAGS,
            ApiField.EXCLUDE_IMAGES_WITH_TAGS,
            ApiField.ENTITIES,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **LabelingJobInfo**.
        """
        return "LabelingJobInfo"

    def __init__(self, api):
        ModuleApi.__init__(self, api)

    def _convert_json_info(self, info: Dict, skip_missing: Optional[bool] = True):
        """ """
        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if field_name in [
                    ApiField.INCLUDE_IMAGES_WITH_TAGS,
                    ApiField.EXCLUDE_IMAGES_WITH_TAGS,
                ]:
                    continue
                value = None
                if type(field_name) is str:
                    if skip_missing is True:
                        value = info.get(field_name, None)
                    else:
                        value = info[field_name]
                elif type(field_name) is tuple:
                    for sub_name in field_name[0]:
                        if value is None:
                            if skip_missing is True:
                                value = info.get(sub_name, None)
                            else:
                                value = info[sub_name]
                        else:
                            value = value[sub_name]
                else:
                    raise RuntimeError("Can not parse field {!r}".format(field_name))

                if field_name == ApiField.FILTER_IMAGES_BY_TAGS:
                    field_values.append(value)
                    include_images_with_tags = []
                    exclude_images_with_tags = []
                    for fv in value:
                        key = ApiField.NAME
                        if key not in fv:
                            key = "title"
                        if fv["positive"] is True:
                            include_images_with_tags.append(fv[key])
                        else:
                            exclude_images_with_tags.append(fv[key])
                    field_values.append(include_images_with_tags)
                    field_values.append(exclude_images_with_tags)
                    continue
                elif (
                    field_name == ApiField.CLASSES_TO_LABEL or field_name == ApiField.TAGS_TO_LABEL
                ):
                    new_value = []
                    for fv in value:
                        key = ApiField.NAME
                        if ApiField.NAME not in fv:
                            key = "title"
                        new_value.append(fv[key])
                    field_values.append(new_value)
                    continue
                elif field_name == ApiField.IMAGES_RANGE:
                    value = (value["start"], value["end"])

                field_values.append(value)

            res = self.InfoType(*field_values)
            return LabelingJobInfo(**res._asdict())

    def _remove_batch_api_method_name(self):
        """Api remove method name."""

        return "jobs.bulk.remove"

    def _remove_batch_field_name(self):
        """Returns onstant string that represents API field name."""

        return ApiField.IDS

    def create(
        self,
        name: str,
        dataset_id: int,
        user_ids: List[int],
        readme: Optional[str] = None,
        description: Optional[str] = None,
        classes_to_label: Optional[List[str]] = None,
        objects_limit_per_image: Optional[int] = None,
        tags_to_label: Optional[List[str]] = None,
        tags_limit_per_image: Optional[int] = None,
        include_images_with_tags: Optional[List[str]] = None,
        exclude_images_with_tags: Optional[List[str]] = None,
        images_range: Optional[List[int, int]] = None,
        reviewer_id: Optional[int] = None,
        images_ids: Optional[List[int]] = [],
    ) -> List[LabelingJobInfo]:
        """
        Creates Labeling Job and assigns given Users to it.

        :param name: Labeling Job name in Supervisely.
        :type name: str
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param user_ids: User IDs in Supervisely to assign Users as labelers to Labeling Job.
        :type user_ids: List[int]
        :param readme: Additional information about Labeling Job.
        :type readme: str, optional
        :param description: Description of Labeling Job.
        :type description: str, optional
        :param classes_to_label: List of classes to label in Dataset.
        :type classes_to_label: List[str], optional
        :param objects_limit_per_image: Limit the number of objects that the labeler can create on each image.
        :type objects_limit_per_image: int, optional
        :param tags_to_label: List of tags to label in Dataset.
        :type tags_to_label: List[str], optional
        :param tags_limit_per_image: Limit the number of tags that the labeler can create on each image.
        :type tags_limit_per_image: int, optional
        :param include_images_with_tags: Include images with given tags for processing by labeler.
        :type include_images_with_tags: List[str], optional
        :param exclude_images_with_tags: Exclude images with given tags for processing by labeler.
        :type exclude_images_with_tags: List[str], optional
        :param images_range: Limit number of images to be labeled for each labeler.
        :type images_range: List[int, int], optional
        :param reviewer_id: User ID in Supervisely to assign User as Reviewer to Labeling Job.
        :type reviewer_id: int, optional
        :param images_ids: List of images ids to label in dataset
        :type images_ids: List[int], optional
        :return: List of information about new Labeling Job. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[LabelingJobInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            user_name = 'alex'
            dataset_id = 602
            new_label_jobs = api.labeling_job.create(user_name, dataset_id, user_ids=[111, 222], readme='Readmy text',
                                                     description='Work for labelers', objects_limit_per_image=5, tags_limit_per_image=3)
            print(new_label_jobs)
            # Output: [
            #     [
            #         92,
            #         "alex (#1) (#3)",
            #         "Readmy text",
            #         "Work for labelers",
            #         13,
            #         29,
            #         "Labelling Workspace",
            #         494,
            #         "Test Dataset",
            #         602,
            #         "ds1",
            #         8,
            #         "alex",
            #         111,
            #         "quantigo273",
            #         8,
            #         "alex",
            #         "2021-03-25T11:04:34.031Z",
            #         null,
            #         null,
            #         "pending",
            #         false,
            #         3,
            #         0,
            #         0,
            #         0,
            #         0,
            #         [],
            #         [],
            #         [
            #             null,
            #             null
            #         ],
            #         5,
            #         3,
            #         [],
            #         [],
            #         [],
            #         [
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287244,
            #                 "name": "IMG_0813"
            #             },
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287246,
            #                 "name": "IMG_0432"
            #             },
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287245,
            #                 "name": "IMG_0315"
            #             }
            #         ]
            #     ],
            #     [
            #         93,
            #         "alex (#2) (#3)",
            #         "Readmy text",
            #         "Work for labelers",
            #         13,
            #         29,
            #         "Labelling Workspace",
            #         494,
            #         "Test Dataset",
            #         602,
            #         "ds1",
            #         8,
            #         "alex",
            #         222,
            #         "quantigo19",
            #         8,
            #         "alex",
            #         "2021-03-25T11:04:34.031Z",
            #         null,
            #         null,
            #         "pending",
            #         false,
            #         3,
            #         0,
            #         0,
            #         0,
            #         0,
            #         [],
            #         [],
            #         [
            #             null,
            #             null
            #         ],
            #         5,
            #         3,
            #         [],
            #         [],
            #         [],
            #         [
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287248,
            #                 "name": "IMG_8454"
            #             },
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287249,
            #                 "name": "IMG_6896"
            #             },
            #             {
            #                 "reviewStatus": "none",
            #                 "id": 287247,
            #                 "name": "IMG_1942"
            #             }
            #         ]
            #     ]
            # ]
        """
        if classes_to_label is None:
            classes_to_label = []
        if tags_to_label is None:
            tags_to_label = []

        filter_images_by_tags = []
        if include_images_with_tags is not None:
            for tag_name in include_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": True})

        if exclude_images_with_tags is not None:
            for tag_name in exclude_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": False})

        if objects_limit_per_image is None:
            objects_limit_per_image = 0

        if tags_limit_per_image is None:
            tags_limit_per_image = 0

        data = {
            ApiField.NAME: name,
            ApiField.DATASET_ID: dataset_id,
            ApiField.USER_IDS: user_ids,
            # ApiField.DESCRIPTION: description,
            ApiField.META: {
                "classes": classes_to_label,
                "projectTags": tags_to_label,
                "imageTags": filter_images_by_tags,
                "imageFiguresLimit": objects_limit_per_image,
                "imageTagsLimit": tags_limit_per_image,
                "entityIds": images_ids,
            },
        }

        if readme is not None:
            data[ApiField.README] = str(readme)

        if description is not None:
            data[ApiField.DESCRIPTION] = str(description)

        if images_range is not None:
            if len(images_range) != 2:
                raise RuntimeError("images_range has to contain 2 elements (start, end)")
            images_range = {"start": images_range[0], "end": images_range[1]}
            data[ApiField.META]["range"] = images_range

        if reviewer_id is not None:
            data[ApiField.REVIEWER_ID] = reviewer_id

        response = self._api.post("jobs.add", data)
        # created_jobs_json = response.json()

        created_jobs = []
        for job in response.json():
            created_jobs.append(self.get_info_by_id(job[ApiField.ID]))
        return created_jobs

    def get_list(
        self,
        team_id: int,
        created_by_id: Optional[int] = None,
        assigned_to_id: Optional[int] = None,
        project_id: Optional[int] = None,
        dataset_id: Optional[int] = None,
        show_disabled: Optional[bool] = False,
    ) -> List[LabelingJobInfo]:
        """
        Get list of information about Labeling Job in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param created_by_id: ID of the User who created the LabelingJob.
        :type created_by_id: int, optional
        :param assigned_to_id: ID of the assigned User.
        :type assigned_to_id: int, optional
        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int, optional
        :param show_disabled: Show disabled Labeling Jobs.
        :type show_disabled: bool, optional
        :return: List of information about Labeling Jobs. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[LabelingJobInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            label_jobs = api.labeling_job.get_list(4)
            print(label_jobs)
            # Output: [
            #     [
            #         2,
            #         "Annotation Job (#1) (#1) (dataset_01)",
            #         "",
            #         "",
            #         4,
            #         8,
            #         "First Workspace",
            #         58,
            #         "tutorial_project",
            #         54,
            #         "dataset_01",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         "2020-04-08T15:10:12.618Z",
            #         "2020-04-08T15:10:19.833Z",
            #         "2020-04-08T15:13:39.788Z",
            #         "completed",
            #         false,
            #         3,
            #         0,
            #         1,
            #         2,
            #         2,
            #         [],
            #         [],
            #         [
            #             1,
            #             5
            #         ],
            #         null,
            #         null,
            #         [],
            #         [],
            #         [],
            #         null
            #     ],
            #     [
            #         3,
            #         "Annotation Job (#1) (#2) (dataset_02)",
            #         "",
            #         "",
            #         4,
            #         8,
            #         "First Workspace",
            #         58,
            #         "tutorial_project",
            #         55,
            #         "dataset_02",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         "2020-04-08T15:10:12.618Z",
            #         "2020-04-08T15:15:46.749Z",
            #         "2020-04-08T15:17:33.572Z",
            #         "completed",
            #         false,
            #         2,
            #         0,
            #         0,
            #         2,
            #         2,
            #         [],
            #         [],
            #         [
            #             1,
            #             5
            #         ],
            #         null,
            #         null,
            #         [],
            #         [],
            #         [],
            #         null
            #     ]
            # ]
        """
        filters = []
        if created_by_id is not None:
            filters.append(
                {"field": ApiField.CREATED_BY_ID[0][0], "operator": "=", "value": created_by_id}
            )
        if assigned_to_id is not None:
            filters.append(
                {"field": ApiField.ASSIGNED_TO_ID[0][0], "operator": "=", "value": assigned_to_id}
            )
        if project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": project_id})
        if dataset_id is not None:
            filters.append({"field": ApiField.DATASET_ID, "operator": "=", "value": dataset_id})
        return self.get_list_all_pages(
            "jobs.list",
            {ApiField.TEAM_ID: team_id, "showDisabled": show_disabled, ApiField.FILTER: filters},
        )

    def stop(self, id: int) -> None:
        """
        Makes Labeling Job unavailable for labeler with given User ID.

        :param id: User ID in Supervisely.
        :type id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.stop(9)
        """
        self._api.post("jobs.stop", {ApiField.ID: id})

    def get_info_by_id(self, id: int) -> LabelingJobInfo:
        """
        Get Labeling Job information by ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Information about Labeling Job. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`LabelingJobInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            label_job_info = api.labeling_job.get_info_by_id(2)
            print(label_job_info)
            # Output: [
            #     2,
            #     "Annotation Job (#1) (#1) (dataset_01)",
            #     "",
            #     "",
            #     4,
            #     8,
            #     "First Workspace",
            #     58,
            #     "tutorial_project",
            #     54,
            #     "dataset_01",
            #     4,
            #     "anna",
            #     4,
            #     "anna",
            #     4,
            #     "anna",
            #     "2020-04-08T15:10:12.618Z",
            #     "2020-04-08T15:10:19.833Z",
            #     "2020-04-08T15:13:39.788Z",
            #     "completed",
            #     false,
            #     3,
            #     0,
            #     1,
            #     2,
            #     2,
            #     [],
            #     [],
            #     [
            #         1,
            #         5
            #     ],
            #     null,
            #     null,
            #     [],
            #     [],
            #     [],
            #     [
            #         {
            #             "reviewStatus": "rejected",
            #             "id": 283,
            #             "name": "image_03"
            #         },
            #         {
            #             "reviewStatus": "accepted",
            #             "id": 282,
            #             "name": "image_02"
            #         },
            #         {
            #             "reviewStatus": "accepted",
            #             "id": 281,
            #             "name": "image_01"
            #         }
            #     ]
            # ]
        """
        return self._get_info_by_id(id, "jobs.info")

    def archive(self, id: int) -> None:
        """
        Archives Labeling Job with given ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.archive(23)
        """
        self._api.post("jobs.archive", {ApiField.ID: id})

    def get_status(self, id: int) -> LabelingJobApi.Status:
        """
        Get status of Labeling Job with given ID.

        :param id: Labeling job ID in Supervisely.
        :type id: int
        :return: Labeling Job Status
        :rtype: :class:`Status<supervisely.api.labeling_job_api.LabelingJobApi.Status>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            job_status = api.labeling_job.get_status(4)
            print(job_status) # pending
        """
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        """ """
        # there is no ERROR status for labeling job
        pass

    def wait(
        self,
        id: int,
        target_status: str,
        wait_attempts: Optional[int] = None,
        wait_attempt_timeout_sec: Optional[int] = None,
    ) -> None:
        """
        Wait for a Labeling Job to change to the expected target status.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param target_status: Expected result status of Labeling Job.
        :type target_status: str
        :param wait_attempts: Number of attempts to retry, when :class:`WaitingTimeExceeded` raises.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Time between attempts.
        :type wait_attempt_timeout_sec: int, optional
        :raises: :class:`WaitingTimeExceeded`, if waiting time exceeded
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.wait(4, 'completed', wait_attempts=2, wait_attempt_timeout_sec=1)
            # supervisely.api.module_api.WaitingTimeExceeded: Waiting time exceeded
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status is target_status:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded("Waiting time exceeded")

    def get_stats(self, id: int) -> Dict:
        """
        Get stats of given Labeling Job ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Dict with information about given Labeling Job
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            status = api.labeling_job.get_stats(3)
            print(status)
            # Output: {
            #     "job": {
            #         "editingDuration": 0,
            #         "annotationDuration": 720,
            #         "id": 3,
            #         "name": "Annotation Job (#1) (#2) (dataset_02)",
            #         "startedAt": "2020-04-08T15:15:46.749Z",
            #         "finishedAt": "2020-04-08T15:17:33.572Z",
            #         "imagesCount": 2,
            #         "finishedImagesCount": 2,
            #         "tagsStats": [
            #             {
            #                 "id": 24,
            #                 "color": "#ED68A1",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "car_color"
            #             },
            #             {
            #                 "id": 19,
            #                 "color": "#A0A08C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "cars_number"
            #             },
            #             {
            #                 "id": 20,
            #                 "color": "#D98F7E",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "like"
            #             },
            #             {
            #                 "id": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "person_gender"
            #             },
            #             {
            #                 "parentId": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "person_gender (male)"
            #             },
            #             {
            #                 "parentId": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "person_gender (female)"
            #             },
            #             {
            #                 "id": 21,
            #                 "color": "#855D79",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "situated"
            #             },
            #             {
            #                 "parentId": 21,
            #                 "color": "#855D79",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "situated (inside)"
            #             },
            #             {
            #                 "parentId": 21,
            #                 "color": "#855D79",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "situated (outside)"
            #             },
            #             {
            #                 "id": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "vehicle_age"
            #             },
            #             {
            #                 "parentId": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "vehicle_age (modern)"
            #             },
            #             {
            #                 "parentId": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "vehicle_age (vintage)"
            #             }
            #         ]
            #     },
            #     "classes": [
            #         {
            #             "id": 43,
            #             "color": "#F6FF00",
            #             "shape": "rectangle",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "bike",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 42,
            #             "color": "#BE55CE",
            #             "shape": "polygon",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "car",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 41,
            #             "color": "#FD0000",
            #             "shape": "polygon",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "dog",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 40,
            #             "color": "#00FF12",
            #             "shape": "bitmap",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "person",
            #             "labelsCount": 0
            #         }
            #     ],
            #     "images": {
            #         "total": 2,
            #         "images": [
            #             {
            #                 "id": 285,
            #                 "reviewStatus": "accepted",
            #                 "annotationDuration": 0,
            #                 "totalDuration": 0,
            #                 "name": "image_01",
            #                 "labelsCount": 0
            #             },
            #             {
            #                 "id": 284,
            #                 "reviewStatus": "accepted",
            #                 "annotationDuration": 0,
            #                 "totalDuration": 0,
            #                 "name": "image_02",
            #                 "labelsCount": 0
            #             }
            #         ]
            #     }
            # }
        """
        response = self._api.post("jobs.stats", {ApiField.ID: id})
        return response.json()

    def get_activity(
        self, team_id: int, job_id: int, progress_cb: Optional[Union[tqdm, Callable]] = None
    ) -> DataFrame:
        import pandas as pd

        """
        Get all activity for given Labeling Job by ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param job_id: Labeling Job ID in Supervisely.
        :type job_id: int
        :param progress_cb: Function for tracking progress
        :type progress_cb: tqdm, optional
        :return: Activity data as `pd.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
        :rtype: :class:`pd.DataFrame`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            activity = api.labeling_job.get_activity(3)
            print(activity)
            # Output:
            #   userId         action  ... tagId                 meta
            # 0       4  update_figure  ...   NaN                   {}
            # 1       4  create_figure  ...   NaN                   {}
            # 2       4     attach_tag  ...  20.0                   {}
            # 3       4     attach_tag  ...  21.0  {'value': 'inside'}
            # 4       4     attach_tag  ...  24.0      {'value': '12'}
            # 5       4  update_figure  ...   NaN                   {}
            # 6       4  update_figure  ...   NaN                   {}
            # 7       4  update_figure  ...   NaN                   {}
            # 8       4  create_figure  ...   NaN                   {}
            # 9       4  update_figure  ...   NaN                   {}
            # [10 rows x 18 columns]
        """
        activity = self._api.team.get_activity(
            team_id, filter_job_id=job_id, progress_cb=progress_cb
        )
        df = pd.DataFrame(activity)
        return df

    def set_status(self, id: int, status: str) -> None:
        """
        Sets Labeling Job status.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param status: New Labeling Job status
        :type status: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.api.labeling_job_api.LabelingJobApi.Status import COMPLETED

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.set_status(id=9, status="completed")
        """
        self._api.post("jobs.set-status", {ApiField.ID: id, ApiField.STATUS: status})

    def get_project_meta(self, id: int) -> ProjectMeta:
        """
        Returns project meta with classes and tags used in the labeling job with given id.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Project meta of the labeling job with given id.
        :rtype: :class:`ProjectMeta`
        """
        job_info = self.get_info_by_id(id)
        project_meta_json = self._api.project.get_meta(job_info.project_id)
        project_meta = ProjectMeta.from_json(project_meta_json)

        job_classes = [
            obj_class
            for obj_class in project_meta.obj_classes
            if obj_class.name in job_info.classes_to_label
        ]
        job_tags = [
            tag_meta
            for tag_meta in project_meta.tag_metas
            if tag_meta.name in job_info.tags_to_label
        ]
        job_meta = ProjectMeta(obj_classes=job_classes, tag_metas=job_tags)
        return job_meta

    def get_annotations(
        self, id: int, image_ids: List[int], project_meta: ProjectMeta = None
    ) -> List[Annotation]:
        """
        Return annotations for given image ids from labeling job with given id.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param image_ids: Image IDs in Supervisely.
        :type image_ids: int
        :param project_meta: Project meta of the labeling job with given id. Can be retrieved with :func:`get_project_meta`.
        :type project_meta: :class:`ProjectMeta`, optional
        :return: Annotation for given image id from labeling job with given id.
        :rtype: :class:`Annotation`
        """

        def _get_geometry(type: str, data: dict):
            if type == "bitmap":
                geometry = Bitmap.from_json(data)
            elif type == "point":
                geometry = Point.from_json(data)
            elif type == "rectangle":
                geometry = Rectangle.from_json(data)
            elif type == "polygon":
                geometry = Polygon.from_json(data)
            elif type == "line":
                geometry = Polyline.from_json(data)
            elif type == "graph":
                geometry = GraphNodes.from_json(data)
            else:
                geometry = None
            return geometry

        def _create_tags_from_labeling_job(
            job_tags_map: List[dict], project_meta: ProjectMeta
        ) -> List[Tag]:
            tags = []
            for tag in job_tags_map:
                tag_meta = project_meta.get_tag_meta_by_id(tag["tagId"])
                if tag_meta is None:
                    continue
                tag_value = tag.get("value", None)
                tag = Tag(tag_meta, value=tag_value)
                tags.append(tag)
            return tags

        def _create_labels_from_labeling_job(
            figures_map: List[FigureInfo], project_meta: ProjectMeta
        ) -> List[Label]:
            labels = []
            for figure in figures_map:
                obj_class = project_meta.get_obj_class_by_id(figure.class_id)
                if obj_class is None:
                    continue
                geometry = _get_geometry(figure.geometry_type, figure.geometry)
                if geometry is None:
                    continue
                tags = _create_tags_from_labeling_job(figure.tags, project_meta)
                label = Label(obj_class=obj_class, geometry=geometry, tags=tags)
                labels.append(label)
            return labels

        self._api.add_header("x-job-id", str(id))
        job_info = self.get_info_by_id(id)
        figures_map = self._api.image.figure.download(job_info.dataset_id, image_ids)
        images = self._api.image.get_list(
            job_info.dataset_id,
            filters=[{ApiField.FIELD: ApiField.ID, "operator": "in", "value": image_ids}],
        )
        self._api.pop_header("x-job-id")

        if project_meta is None:
            project_meta = self.get_project_meta(id)

        anns = []
        for image in images:
            img_figures = figures_map.get(image.id, [])
            img_tags = _create_tags_from_labeling_job(image.tags, project_meta)
            labels = _create_labels_from_labeling_job(img_figures, project_meta)
            ann = Annotation(img_size=(image.height, image.width), labels=labels, img_tags=img_tags)
            anns.append(ann)
        return anns



import json
import time
import requests
from typing import List, Optional, Union, Iterator, Dict, Any, Tuple

try:
    from typing import Literal
except ImportError:
    # for compatibility with python 3.7
    from typing_extensions import Literal
import yaml
from requests import Timeout, HTTPError

import supervisely as sly
from supervisely.io.network_exceptions import process_requests_exception
from supervisely.sly_logger import logger


class SessionJSON:
    def __init__(
        self,
        api: sly.Api,
        task_id: int = None,
        session_url: str = None,
        inference_settings: Union[dict, str] = None,
    ):
        """
        A convenient class for inference of deployed models.
        This class will return raw JSON predictions (dict).
        If you want to work with `sly.Annotation` format, please use `sly.nn.inference.Session` class.

        You need first to serve NN model you want and get its `task_id`.
        The `task_id` can be obtained from the Supervisely platform, going to `START` -> `App sessions` page.

        Note: Either a `task_id` or a `session_url` has to be passed as a parameter (not both).

        :param api: initialized :class:`sly.Api` object.
        :type api: sly.Api
        :param task_id: the task_id of a served model in the Supervisely platform. If None, the `session_url` will be used instead, defaults to None
        :type task_id: int, optional
        :param session_url: the url for direct connection to the served model. If None, the `task_id` will be used instead, defaults to None
        :type session_url: str, optional
        :param inference_settings: a dict or a path to YAML file with settings, defaults to None
        :type inference_settings: Union[dict, str], optional


        :Usage example:
         .. code-block:: python
            task_id = 27001
            session = sly.nn.inference.SessionJSON(
                api,
                task_id=task_id,
            )
            print(session.get_session_info())

            image_id = 17551748
            pred = session.inference_image_id(image_id)
            predicted_annotation = sly.Annotation.from_json(pred["annotation"], model_meta)

        """
        assert not (
            task_id is None and session_url is None
        ), "Either `task_id` or `session_url` must be passed."
        assert (
            task_id is None or session_url is None
        ), "Either `task_id` or `session_url` must be passed (not both)."

        self.api = api
        self._task_id = task_id

        if task_id is not None:
            try:
                # TODO: api.task.get_info_by_id get stuck if the task_id isn't exists on the platform
                # https://github.com/supervisely/issues/issues/1873
                task_info = api.task.get_info_by_id(task_id)
            except HTTPError:
                raise ValueError(
                    f"Can't connect to the model. Check if the task_id {task_id} is correct."
                )
        if task_id is not None:
            self._base_url = f'{self.api.server_address}/net/{task_info["meta"]["sessionToken"]}'
        else:
            self._base_url = session_url
        self.set_inference_settings(inference_settings)

        self._session_info = None
        self._default_inference_settings = None
        self._model_meta = None
        self._async_inference_uuid = None
        self._stop_async_inference_flag = False

        # Check connection:
        try:
            self.get_session_info()
        except HTTPError:
            if task_id is not None:
                raise ValueError(
                    f"Can't connect to the model. Check if the task_id {self._task_id} is correct."
                )
            else:
                raise ValueError(
                    f'Can\'t connect to the model. Check if the session_url "{session_url}" is correct.'
                )

    @property
    def task_id(self) -> int:
        return self._task_id

    @property
    def base_url(self) -> str:
        return self._base_url

    def get_session_info(self) -> Dict[str, Any]:
        if self._session_info is None:
            self._session_info = self._get_from_endpoint("get_session_info")
        return self._session_info

    def get_human_readable_info(self, replace_none_with: Optional[str] = None):
        hr_info = {}
        info = self.get_session_info()

        for name, data in info.items():
            hr_name = name.replace("_", " ").capitalize()
            if data is None:
                hr_info[hr_name] = replace_none_with
            else:
                hr_info[hr_name] = data

        return hr_info

    def get_default_inference_settings(self) -> Dict[str, Any]:
        if self._default_inference_settings is None:
            resp = self._get_from_endpoint("get_custom_inference_settings")
            settings = resp["settings"]
            if isinstance(settings, str):
                settings = yaml.safe_load(settings)
            self._default_inference_settings = settings
        return self._default_inference_settings

    def get_model_meta(self) -> Dict[str, Any]:
        if self._model_meta is None:
            meta_json = self._get_from_endpoint("get_output_classes_and_tags")
            self._model_meta = meta_json
        return self._model_meta

    def update_inference_settings(self, **inference_settings) -> Dict[str, Any]:
        self.inference_settings.update(inference_settings)
        return self.inference_settings

    def set_inference_settings(self, inference_settings) -> Dict[str, Any]:
        self._set_inference_settings_dict_or_yaml(inference_settings)
        return self.inference_settings

    def _set_inference_settings_dict_or_yaml(self, dict_or_yaml_path) -> None:
        if isinstance(dict_or_yaml_path, str):
            with open(dict_or_yaml_path, "r") as f:
                self.inference_settings: dict = yaml.safe_load(f)
        elif isinstance(dict_or_yaml_path, dict):
            self.inference_settings = dict_or_yaml_path
        else:
            self.inference_settings = {}

    def inference_image_id(self, image_id: int) -> Dict[str, Any]:
        endpoint = "inference_image_id"
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        json_body["state"]["image_id"] = image_id
        resp = self._post(url, json=json_body)
        return resp.json()

    def inference_image_ids(self, image_ids: List[int]) -> List[Dict[str, Any]]:
        endpoint = "inference_batch_ids"
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        json_body["state"]["batch_ids"] = image_ids
        resp = self._post(url, json=json_body)
        return resp.json()

    def inference_image_url(self, url: str) -> Dict[str, Any]:
        endpoint = "inference_image_url"
        endpoint_url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        json_body["state"]["image_url"] = url
        resp = self._post(endpoint_url, json=json_body)
        return resp.json()

    def inference_image_path(self, image_path: str) -> Dict[str, Any]:
        endpoint = "inference_image"
        url = f"{self._base_url}/{endpoint}"
        opened_file = open(image_path, "rb")
        settings_json = json.dumps({"settings": self.inference_settings})
        uploads = [
            ("files", opened_file),
            ("settings", (None, settings_json, "text/plain")),
        ]
        resp = self._post(url, files=uploads)
        opened_file.close()
        return resp.json()

    def inference_image_paths(self, image_paths: List[str]) -> List[Dict[str, Any]]:
        endpoint = "inference_batch"
        url = f"{self._base_url}/{endpoint}"
        files = [("files", open(f, "rb")) for f in image_paths]
        settings_json = json.dumps({"settings": self.inference_settings})
        uploads = files + [("settings", (None, settings_json, "text/plain"))]
        resp = self._post(url, files=uploads)
        for _, f in files:
            f.close()
        return resp.json()

    def inference_video_id(
        self,
        video_id: int,
        start_frame_index: int = None,
        frames_count: int = None,
        frames_direction: Literal["forward", "backward"] = None,
    ) -> Dict[str, Any]:
        endpoint = "inference_video_id"
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        state = json_body["state"]
        state["videoId"] = video_id
        state.update(
            self._collect_state_for_infer_video(start_frame_index, frames_count, frames_direction)
        )
        resp = self._post(url, json=json_body)
        return resp.json()

    def inference_video_id_async(
        self,
        video_id: int,
        start_frame_index: int = None,
        frames_count: int = None,
        frames_direction: Literal["forward", "backward"] = None,
        process_fn=None,
        preparing_cb=None,
    ) -> Iterator:
        if self._async_inference_uuid:
            logger.info(
                "Trying to run a new inference while `_async_inference_uuid` already exists. Stopping the old one..."
            )
            try:
                self.stop_async_inference()
                self._on_async_inference_end()
            except Exception as exc:
                logger.error(f"An error has occurred while stopping the previous inference. {exc}")
        endpoint = "inference_video_id_async"
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        state = json_body["state"]
        state["videoId"] = video_id
        state.update(
            self._collect_state_for_infer_video(start_frame_index, frames_count, frames_direction)
        )
        resp = self._post(url, json=json_body).json()
        self._async_inference_uuid = resp["inference_request_uuid"]
        self._stop_async_inference_flag = False

        current = 0
        prev_current = 0
        if preparing_cb:
            # wait for inference status
            resp = self._get_preparing_progress()
            while resp.get("status") is None:
                time.sleep(2)
                resp = self._get_preparing_progress()

            if resp["status"] == "download_video":
                progress_widget = preparing_cb(
                    message="Downloading Video",
                    total=resp["total"],
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                )
            elif resp["status"] == "download_frames":
                progress_widget = preparing_cb(message="Downloading Frames", total=resp["total"])

            while resp["status"] in ["download_video", "download_frames"]:
                resp = self._get_preparing_progress()
                current = resp["current"]
                progress_widget.update(current - prev_current)
                prev_current = current

            resp = self._get_preparing_progress()
            if resp["status"] == "cut":
                progress_widget = preparing_cb(message="Cutting frames", total=resp["total"])
            while resp["status"] == "cut":
                resp = self._get_preparing_progress()
                current = resp["current"]
                progress_widget.update(current - prev_current)
                prev_current = current

        logger.info("Inference has started:", extra=resp)
        resp, has_started = self._wait_for_async_inference_start()
        frame_iterator = AsyncInferenceIterator(
            resp["progress"]["total"], self, process_fn=process_fn
        )
        return frame_iterator

    def stop_async_inference(self) -> Dict[str, Any]:
        endpoint = "stop_inference"
        resp = self._get_from_endpoint_for_async_inference(endpoint)
        self._stop_async_inference_flag = True
        logger.info("Inference will be stopped on the server")
        return resp

    def _get_inference_progress(self) -> Dict[str, Any]:
        endpoint = "get_inference_progress"
        return self._get_from_endpoint_for_async_inference(endpoint)

    def _get_preparing_progress(self) -> Dict[str, Any]:
        endpoint = "get_preparing_progress"
        return self._get_from_endpoint_for_async_inference(endpoint)

    def _pop_pending_results(self) -> Dict[str, Any]:
        endpoint = "pop_inference_results"
        return self._get_from_endpoint_for_async_inference(endpoint)

    def _clear_inference_request(self) -> Dict[str, Any]:
        endpoint = "clear_inference_request"
        return self._get_from_endpoint_for_async_inference(endpoint)

    def _wait_for_async_inference_start(self, delay=1, timeout=None) -> Tuple[dict, bool]:
        logger.info("The video is preparing on the server, this may take a while...")
        has_started = False
        timeout_exceeded = False
        t0 = time.time()
        while not has_started and not timeout_exceeded:
            resp = self._get_inference_progress()
            has_started = bool(resp["result"]) or resp["progress"]["total"] != 1
            if not has_started:
                time.sleep(delay)
            timeout_exceeded = timeout and time.time() - t0 > timeout
        if timeout_exceeded:
            self.stop_async_inference()
            self._on_async_inference_end()
            raise Timeout("Timeout exceeded. The server didn't start the inference")
        return resp, has_started

    def _wait_for_new_pending_results(self, delay=1, timeout=600) -> List[dict]:
        logger.debug("waiting pending results...")
        has_results = False
        timeout_exceeded = False
        t0 = time.time()
        while not has_results and not timeout_exceeded:
            resp = self._pop_pending_results()
            pending_results = resp["pending_results"]
            has_results = bool(pending_results)
            if resp["is_inferring"] is False:
                break
            if not has_results:
                time.sleep(delay)
            timeout_exceeded = timeout and time.time() - t0 > timeout
        if timeout_exceeded:
            self.stop_async_inference()
            self._on_async_inference_end()
            raise Timeout("Timeout exceeded. Pending results not received from the server.")
        if len(pending_results) == 0 and resp["is_inferring"]:
            logger.warn(
                "The model is inferring yet, but new pending results have not received from the serving app. "
                "This may lead to not all samples will be inferred."
            )
        return pending_results

    def _on_async_inference_end(self):
        logger.debug("callback: on_async_inference_end()")
        try:
            self._clear_inference_request()
        finally:
            self._async_inference_uuid = None

    def _post(self, *args, retries=5, **kwargs) -> requests.Response:
        url = kwargs.get("url") or args[0]
        method = url[len(self._base_url) :]
        for retry_idx in range(retries):
            try:
                response = requests.post(*args, **kwargs)
                if response.status_code != requests.codes.ok:  # pylint: disable=no-member
                    sly.Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(
                    logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=5,
                    retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                )
                if retry_idx + 1 == retries:
                    raise exc

    def _get_from_endpoint(self, endpoint) -> Dict[str, Any]:
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body()
        resp = self._post(url, json=json_body)
        return resp.json()

    def _get_from_endpoint_for_async_inference(self, endpoint) -> Dict[str, Any]:
        url = f"{self._base_url}/{endpoint}"
        json_body = self._get_default_json_body_for_async_inference()
        resp = self._post(url, json=json_body)
        return resp.json()

    def _collect_state_for_infer_video(
        self,
        start_frame_index: int = None,
        frames_count: int = None,
        frames_direction: Literal["forward", "backward"] = None,
    ) -> Dict[str, Any]:
        state = {}
        if start_frame_index is not None:
            state["startFrameIndex"] = start_frame_index
        if frames_count is not None:
            state["framesCount"] = frames_count
        if frames_direction is not None:
            state["framesDirection"] = frames_direction
        return state

    def _get_default_json_body(self) -> Dict[str, Any]:
        return {
            "state": {"settings": self.inference_settings},
            "context": {},
            "api_token": self.api.token,
        }

    def _get_default_json_body_for_async_inference(self) -> Dict[str, Any]:
        json_body = self._get_default_json_body()
        json_body["state"]["inference_request_uuid"] = self._async_inference_uuid
        return json_body


class AsyncInferenceIterator:
    def __init__(self, total, nn_api: SessionJSON, process_fn=None):
        self.total = total
        self.nn_api = nn_api
        self.results_queue = []
        self.process_fn = process_fn

    def __len__(self) -> int:
        return self.total

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Dict[str, Any]:
        try:
            if self.nn_api._stop_async_inference_flag:
                raise StopIteration
            if not self.results_queue:
                pending_results = self.nn_api._wait_for_new_pending_results()
                self.results_queue += pending_results
            if not self.results_queue:
                raise StopIteration

        except StopIteration as stop_iteration:
            self.nn_api._on_async_inference_end()
            raise stop_iteration
        except Exception as ex:
            # Any exceptions will be handled here
            self.nn_api.stop_async_inference()
            self.nn_api._on_async_inference_end()
            raise ex

        pred = self.results_queue.pop(0)
        if self.process_fn is not None:
            return self.process_fn(pred)
        else:
            return pred


class Session(SessionJSON):
    def __init__(
        self,
        api: sly.Api,
        task_id: int = None,
        session_url: str = None,
        inference_settings: Union[dict, str] = None,
    ):
        """
        A convenient class for inference of deployed models.
        This class will return predictions in the `sly.Annotation` format.
        If you want to work with raw JSON dicts, please use `sly.nn.inference.SessionJSON` class.

        You need first to serve NN model you want and get its `task_id`.
        The `task_id` can be obtained from the Supervisely platform, going to `START` -> `App sessions` page.

        Note: Either a `task_id` or a `session_url` has to be passed as a parameter (not both).

        :param api: initialized :class:`sly.Api` object.
        :type api: sly.Api
        :param task_id: the task_id of a served model in the Supervisely platform. If None, the `session_url` will be used instead, defaults to None
        :type task_id: int, optional
        :param session_url: the url for direct connection to the served model. If None, the `task_id` will be used instead, defaults to None
        :type session_url: str, optional
        :param inference_settings: a dict or a path to YAML file with settings, defaults to None
        :type inference_settings: Union[dict, str], optional


        :Usage example:
         .. code-block:: python
            task_id = 27001
            session = sly.nn.inference.Session(
                api,
                task_id=task_id,
            )
            print(session.get_session_info())

            image_id = 17551748
            predicted_annotation = session.inference_image_id(image_id)

        """
        super().__init__(api, task_id, session_url, inference_settings)

    def get_model_meta(self) -> sly.ProjectMeta:
        if not isinstance(self._model_meta, sly.ProjectMeta):
            model_meta_json = super().get_model_meta()
            model_meta = sly.ProjectMeta.from_json(model_meta_json)
            self._model_meta = model_meta
        return self._model_meta

    def inference_image_id(self, image_id: int) -> sly.Annotation:
        pred_json = super().inference_image_id(image_id)
        pred_ann = self._convert_to_sly_annotation(pred_json)
        return pred_ann

    def inference_image_path(self, image_path: str) -> sly.Annotation:
        pred_json = super().inference_image_path(image_path)
        pred_ann = self._convert_to_sly_annotation(pred_json)
        return pred_ann

    def inference_image_url(self, url: str) -> sly.Annotation:
        pred_json = super().inference_image_url(url)
        pred_ann = self._convert_to_sly_annotation(pred_json)
        return pred_ann

    def inference_image_ids(self, image_ids: List[int]) -> List[sly.Annotation]:
        pred_list_raw = super().inference_image_ids(image_ids)
        predictions = self._convert_to_sly_annotation_batch(pred_list_raw)
        return predictions

    def inference_image_paths(self, image_paths: List[str]) -> List[sly.Annotation]:
        pred_list_raw = super().inference_image_paths(image_paths)
        predictions = self._convert_to_sly_annotation_batch(pred_list_raw)
        return predictions

    def inference_video_id(
        self,
        video_id: int,
        start_frame_index: int = None,
        frames_count: int = None,
        frames_direction: Literal["forward", "backward"] = None,
    ) -> List[sly.Annotation]:
        pred_list_raw = super().inference_video_id(
            video_id, start_frame_index, frames_count, frames_direction
        )
        pred_list_raw = pred_list_raw["ann"]
        predictions = self._convert_to_sly_annotation_batch(pred_list_raw)
        return predictions

    def inference_video_id_async(
        self,
        video_id: int,
        start_frame_index: int = None,
        frames_count: int = None,
        frames_direction: Literal["forward", "backward"] = None,
    ) -> AsyncInferenceIterator:
        frame_iterator = super().inference_video_id_async(
            video_id,
            start_frame_index,
            frames_count,
            frames_direction,
            process_fn=self._convert_to_sly_annotation,
        )
        return frame_iterator

    def _convert_to_sly_annotation(self, pred_json: dict):
        model_meta = self.get_model_meta()
        pred_ann = sly.Annotation.from_json(pred_json["annotation"], model_meta)
        return pred_ann

    def _convert_to_sly_annotation_batch(self, pred_list_raw: List[dict]):
        model_meta = self.get_model_meta()
        predictions = [
            sly.Annotation.from_json(pred["annotation"], model_meta) for pred in pred_list_raw
        ]
        return predictions



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class TimePicker(Widget):
    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: str = "asd",
        start: str = "09:00",
        step: str = "00:15",
        end: str = "22:30",
        placeholder: str = None,
        size: Literal["large", "small", "mini"] = None,
        readonly: bool = False,
        editable: bool = True,
        clearable: bool = True,
        widget_id: str = None,
    ):
        self._validate_value(value)
        self._value = value
        self._start = start
        self._step = step
        self._end = end
        self._placeholder = placeholder
        self._size = size
        self._readonly = readonly
        self._editable = editable
        self._clearable = clearable

        self._changes_handled = False
        # self._value = None
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "placeholder": self._placeholder,
            "size": self._size,
            "readonly": self._readonly,
            "editable": self._editable,
            "clearable": self._clearable,
            "picker_options": {"start": self._start, "step": self._step, "end": self._end},
        }

    def get_json_state(self):
        return {"value": self._value}

    def _validate_value(self, value):
        if value is None or value == "":
            return
        if not isinstance(value, str):
            raise TypeError("value must be str type")

        val_split = value.split(":")
        if len(val_split) != 2:
            raise ValueError("value must be in format 'hh:mm'")
        hh, mm = val_split
        if len(hh) != 2 or len(mm) != 2:
            raise ValueError("value must be in format 'hh:mm'")
        if not hh[0].isdigit() or not hh[1].isdigit() or not mm[0].isdigit() or not mm[1].isdigit():
            raise ValueError("value must be in format 'hh:mm'")

    def set_value(self, value: str):
        self._validate_value(value)
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def get_picker_options(self):
        return DataJson()[self.widget_id]["picker_options"]

    def set_start(self, value: str):
        self._start = value
        DataJson()[self.widget_id]["picker_options"]["start"] = self._start
        DataJson().send_changes()

    def set_end(self, value: str):
        self._end = value
        DataJson()[self.widget_id]["picker_options"]["end"] = self._end
        DataJson().send_changes()

    def set_step(self, value: str):
        self._step = value
        DataJson()[self.widget_id]["picker_options"]["step"] = self._step
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(TimePicker.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._value = res
            func(res)

        return _click



# coding: utf-8

# docs
from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class FigureInfo(NamedTuple):
    id: int
    class_id: int
    updated_at: str
    created_at: str
    entity_id: int
    object_id: int
    project_id: int
    dataset_id: int
    frame_index: int
    geometry_type: str
    geometry: dict
    tags: list
    meta: dict


class FigureApi(RemoveableBulkModuleApi):
    """
    Figure object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`.
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple FigureInfo information about Figure.

        :Example:

         .. code-block:: python

            FigureInfo(id=588801373,
                       updated_at='2020-12-22T06:37:13.183Z',
                       created_at='2020-12-22T06:37:13.183Z',
                       entity_id=186648101,
                       object_id=112482,
                       project_id=110366,
                       dataset_id=419886,
                       frame_index=0,
                       geometry_type='bitmap',
                       geometry={'bitmap': {'data': 'eJwdlns8...Cgj4=', 'origin': [335, 205]}})
        """
        return [
            ApiField.ID,
            ApiField.CLASS_ID,
            ApiField.UPDATED_AT,
            ApiField.CREATED_AT,
            ApiField.ENTITY_ID,
            ApiField.OBJECT_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.FRAME_INDEX,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY,
            ApiField.TAGS,
            ApiField.META,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.video.figure.info_tuple_name()
            print(tuple_name) # FigureInfo
        """

        return "FigureInfo"

    def get_info_by_id(self, id: int) -> FigureInfo:
        """
        Get Figure information by ID.

        :param id: Figure ID in Supervisely.
        :type id: int
        :return: Information about Figure. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_id = 588801373

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            figure_info = api.video.figure.get_info_by_id(figure_id)
            print(figure_info)
            # Output: [
            #     588801373,
            #     "2020-12-22T06:37:13.183Z",
            #     "2020-12-22T06:37:13.183Z",
            #     186648101,
            #     112482,
            #     110366,
            #     419886,
            #     0,
            #     "bitmap",
            #     {
            #         "bitmap": {
            #             "data": "eJw...Cgj4=",
            #             "origin": [
            #                 335,
            #                 205
            #             ]
            #         }
            #     }
            # ]
        """
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometryType",
            "geometry",
            "tags",
            "meta",
        ]
        return self._get_info_by_id(id, "figures.info", {ApiField.FIELDS: fields})

    def create(
        self,
        entity_id: int,
        object_id: int,
        meta: Dict,
        geometry_json: Dict,
        geometry_type: str,
        track_id: int = None,
    ) -> int:
        """"""
        input_figure = {
            ApiField.META: meta,
            ApiField.OBJECT_ID: object_id,
            ApiField.GEOMETRY_TYPE: geometry_type,
            ApiField.GEOMETRY: geometry_json,
        }

        if track_id is not None:
            input_figure[ApiField.TRACK_ID] = track_id

        body = {ApiField.ENTITY_ID: entity_id, ApiField.FIGURES: [input_figure]}

        response = self._api.post("figures.bulk.add", body)
        return response.json()[0][ApiField.ID]

    def get_by_ids(self, dataset_id: int, ids: List[int]) -> List[FigureInfo]:
        """
        Get Figures information by IDs from given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param ids: List of Figures IDs.
        :type ids: List[int]
        :return: List of information about Figures. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 466642
            figures_ids = [642155547, 642155548, 642155549]
            figures_infos = api.video.figure.get_by_ids(dataset_id, figures_ids)
            print(figures_infos)
            # Output: [
            #     [
            #         642155547,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         0,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2240,
            #                         1041
            #                     ],
            #                     [
            #                         2463,
            #                         1187
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155548,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         1,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2248,
            #                         1048
            #                     ],
            #                     [
            #                         2455,
            #                         1176
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155549,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         2,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2237,
            #                         1046
            #                     ],
            #                     [
            #                         2464,
            #                         1179
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ]
            # ]
        """
        filters = [{"field": "id", "operator": "in", "value": ids}]
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometryType",
            "geometry",
            "tags",
            "meta",
        ]
        figures_infos = self.get_list_all_pages(
            "figures.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters,
                ApiField.FIELDS: fields,
            },
        )

        if len(ids) != len(figures_infos):
            ids_downloaded = [info.id for info in figures_infos]
            raise RuntimeError(
                "Ids don't exist on server: {}".format(set(ids_downloaded) - set(ids))
            )

        id_to_item = {info.id: info for info in figures_infos}

        figures = []
        for input_id in ids:
            figures.append(id_to_item[input_id])

        return figures

    def _append_bulk(
        self,
        entity_id,
        figures_json,
        figures_keys,
        key_id_map: KeyIdMap,
        field_name=ApiField.ENTITY_ID,
    ):
        """"""
        if len(figures_json) == 0:
            return
        for batch_keys, batch_jsons in zip(
            batched(figures_keys, batch_size=100), batched(figures_json, batch_size=100)
        ):
            resp = self._api.post(
                "figures.bulk.add",
                {field_name: entity_id, ApiField.FIGURES: batch_jsons},
            )
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)

    def download(self, dataset_id: int, image_ids: List[int] = None) -> Dict[int, List[FigureInfo]]:
        """
        Method returns dictionary with image ids and list of FigureInfo for given dataset ID. Can be filtered by image IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: List of image IDs within given dataset ID.
        :type image_ids: List[int], optional
        :return: List of information about Figures. See :class:`FigureInfo<FigureInfo>`
        :rtype: :class:`Dict[int, List[FigureInfo]]`
        """
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometryType",
            "geometry",
            "tags",
            "meta",
        ]

        if image_ids is None:
            filters = []
        else:
            filters = [{ApiField.FIELD: ApiField.ENTITY_ID, "operator": "in", "value": image_ids}]
        data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FIELDS: fields,
            ApiField.FILTER: filters,
        }
        resp = self._api.post("figures.list", data)
        infos = resp.json()
        images_figures = {}
        for info in infos["entities"]:
            figure_info = self._convert_json_info(info, True)
            if figure_info.entity_id in images_figures:
                images_figures[figure_info.entity_id].append(figure_info)
            else:
                images_figures[figure_info.entity_id] = [figure_info]
        return images_figures

    def _convert_json_info(self, info: dict, skip_missing=False):
        return super()._convert_json_info(info, True)



# coding: utf-8
from __future__ import annotations
from typing import List, Dict, Optional
import uuid
from uuid import UUID
from bidict import bidict
from supervisely.io.json import dump_json_file, load_json_file

TAGS = "tags"
OBJECTS = "objects"
FIGURES = "figures"
VIDEOS = "videos"

ALLOWED_KEY_TYPES = [TAGS, OBJECTS, VIDEOS, FIGURES]


# @TODO: reimplement to support different item types - videos, volumes, 3d episodes, ...
class KeyIdMap:
    """
    KeyIdMap object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`. It consist from dict with bidict values.

    :Usage example:

     .. code-block:: python

        key_id_map = KeyIdMap()
        print(key_id_map.to_dict())
        # Output: {
        #     "tags": {},
        #     "objects": {},
        #     "figures": {},
        #     "videos": {}
        # }
    """

    def __init__(self):
        self._data = dict()
        self._data[TAGS] = bidict()
        self._data[OBJECTS] = bidict()
        self._data[FIGURES] = bidict()
        self._data[VIDEOS] = bidict()

    def _add(self, key_type, key: uuid.UUID, id: Optional[int] = None):
        """
        Add given data in self._data dictionary. Raise error if data type of any parameter is invalid
        :param key_type: str
        :param key: uuid class object
        :param id: int
        """
        if key_type not in ALLOWED_KEY_TYPES:
            raise RuntimeError(
                "Key type {!r} is not allowed. Allowed types are {}".format(
                    key_type, ALLOWED_KEY_TYPES
                )
            )
        if type(key) is not uuid.UUID:
            raise RuntimeError("Key should be of type uuid.UUID")
        if id is not None and type(id) is not int:
            raise RuntimeError("Id should be of type int")
        self._data[key_type].update(bidict({key: id}))

    def add_object(self, key: UUID, id: int) -> None:
        """
        Add UUID and ID of :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` in KeyIdMap.

        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` ID.
        :type id: int
        :return: :class:`None<None>`
        :rtype: :class:`NoneType<NoneType>`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4() # "0c0033c5b4834d4cbabece4317295f07"
            key_id_map.add_object(new_uuid, 1)
            print(key_id_map.to_dict())
            # Output: {
            #     "tags": {},
            #     "objects": {
            #         "0c0033c5b4834d4cbabece4317295f07": 1
            #     },
            #     "figures": {},
            #     "videos": {}
            # }
        """
        self._add(OBJECTS, key, id)

    def add_tag(self, key: UUID, id: int) -> None:
        """
        Add UUID and ID of :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` in KeyIdMap.

        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` ID.
        :type id: int
        :return: :class:`None<None>`
        :rtype: :class:`NoneType<NoneType>`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4() # "697d005df2a94bb386188c78a61b0a86"
            key_id_map.add_tag(new_uuid, 34)
            print(key_id_map.to_dict())
            # Output: {
            #     "tags": {
            #         "697d005df2a94bb386188c78a61b0a86": 34
            #     },
            #     "objects": {},
            #     "figures": {},
            #     "videos": {}
            # }
        """
        self._add(TAGS, key, id)

    def add_figure(self, key: UUID, id: int) -> None:
        """
        Add UUID and ID of :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` in KeyIdMap.

        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` ID.
        :type id: int
        :return: :class:`None<None>`
        :rtype: :class:`NoneType<NoneType>`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4() # "ac1018e6673d405590086063af8184ca"
            key_id_map.add_figure(new_uuid, 55)
            print(key_id_map.to_dict())
            # Output: {
            #     "tags": {},
            #     "objects": {},
            #     "figures": {
            #         "ac1018e6673d405590086063af8184ca": 55
            #     },
            #     "videos": {}
            # }
        """
        self._add(FIGURES, key, id)

    def add_video(self, key: UUID, id: int) -> None:
        """
        Add UUID and ID of :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` in KeyIdMap.

        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` ID.
        :type id: int
        :return: :class:`None<None>`
        :rtype: :class:`NoneType<NoneType>`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4() # "775f2c581cec44ca8c10419c20c52fcc"
            key_id_map.add_video(new_uuid, 77)
            print(key_id_map.to_dict())
            # Output: {
            #     "tags": {},
            #     "objects": {},
            #     "figures": {},
            #     "videos": {
            #         "775f2c581cec44ca8c10419c20c52fcc": 77
            #     }
            # }
        """
        self._add(VIDEOS, key, id)

    def _get_id_by_key(self, key_type, key: uuid.UUID):
        """
        :param key_type: str
        :param key: uuid class object
        :return: Id by given key. None if there is no such key. Raise error if key type is not uuid.UUID
        """
        if type(key) is not uuid.UUID:
            raise RuntimeError("Key should be of type uuid.UUID")

        if key in self._data[key_type]:
            return self._data[key_type][key]
        else:
            return None

    def _get_key_by_id(self, key_type, id: int):
        """
        :param key_type: str
        :param id: int
        :return: Key by given id. None if there is no such id. Raise error if id type is not int
        """
        if type(id) is not int:
            raise RuntimeError("Id should be of type int")
        if id not in self._data[key_type].inverse:
            return None
        return self._data[key_type].inverse[id]

    def get_object_id(self, key: UUID) -> int:
        """
        Get :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` ID.

        :param key: UUID object.
        :type key: UUID
        :return: :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` ID
        :rtype: :class:`int`

        :Usage example:

         .. code-block:: python

            obj_uuid = '0c0033c5b4834d4cbabece4317295f07'
            obj_id = key_id_map.get_object_id(obj_uuid) # 1
        """
        return self._get_id_by_key(OBJECTS, key)

    def get_tag_id(self, key: UUID) -> int:
        """
        Get :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` ID.

        :param key: UUID object.
        :type key: UUID
        :return: :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` ID
        :rtype: :class:`int`

        :Usage example:

         .. code-block:: python

            tag_uuid = '697d005df2a94bb386188c78a61b0a86'
            tag_id = key_id_map.get_tag_id(tag_uuid) # 34
        """
        return self._get_id_by_key(TAGS, key)

    def get_figure_id(self, key: UUID) -> int:
        """
        Get :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` ID.

        :param key: UUID object.
        :type key: UUID
        :return: :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` ID
        :rtype: :class:`int`

        :Usage example:

         .. code-block:: python

            figure_uuid = 'ac1018e6673d405590086063af8184ca'
            figure_id = key_id_map.get_figure_id(figure_uuid) # 55
        """
        return self._get_id_by_key(FIGURES, key)

    def get_video_id(self, key: UUID) -> int:
        """
        Get :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` ID.

        :param key: UUID object.
        :type key: UUID
        :return: :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` ID
        :rtype: :class:`int`

        :Usage example:

         .. code-block:: python

            video_uuid = '775f2c581cec44ca8c10419c20c52fcc'
            video_id = key_id_map.get_video_id(video_uuid) # 77
        """
        return self._get_id_by_key(VIDEOS, key)

    def get_object_key(self, id: int) -> UUID:
        """
        Get :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` UUID key.

        :param key: :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` ID.
        :type key: int
        :return: :class:`UUID` object
        :rtype: :class:`UUID`

        :Usage example:

         .. code-block:: python

            obj_id = 1
            obj_uuid = key_id_map.get_object_id(obj_id) # '0c0033c5b4834d4cbabece4317295f07'
        """
        return self._get_key_by_id(OBJECTS, id)

    def get_tag_key(self, id: int) -> UUID:
        """
        Get :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` UUID key.

        :param key: :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` ID.
        :type key: int
        :return: :class:`UUID` object
        :rtype: :class:`UUID`

        :Usage example:

         .. code-block:: python

            tag_id = 34
            tag_uuid = key_id_map.get_tag_key(tag_id) # '697d005df2a94bb386188c78a61b0a86'
        """
        return self._get_key_by_id(TAGS, id)

    def get_figure_key(self, id: int) -> UUID:
        """
        Get :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` UUID key.

        :param key: :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` ID.
        :type key: int
        :return: :class:`UUID` object
        :rtype: :class:`UUID`

        :Usage example:

         .. code-block:: python

            figure_id = 55
            figure_uuid = key_id_map.get_figure_key(figure_id) # 'ac1018e6673d405590086063af8184ca'
        """
        return self._get_key_by_id(FIGURES, id)

    def get_video_key(self, id: int) -> UUID:
        """
        Get :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` UUID key.

        :param key: :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` ID.
        :type key: int
        :return: :class:`UUID` object
        :rtype: :class:`UUID`

        :Usage example:

         .. code-block:: python

            video_id = 77
            video_uuid = key_id_map.get_video_key(video_id) # '775f2c581cec44ca8c10419c20c52fcc'
        """
        return self._get_key_by_id(VIDEOS, id)

    def to_dict(self) -> Dict[str, Dict]:
        """
        Convert the KeyIdMap to a dict(bidict values to dictionary with dict values).

        :return: Json format as a dict
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            print(key_id_map.to_dict())
            # Output: {
            #     "tags": {},
            #     "objects": {},
            #     "figures": {},
            #     "videos": {}
            # }
        """
        simple_dict = {}
        for type_str, value_bidict in self._data.items():
            sub_dict = {}
            for uuid, int_id in value_bidict.items():
                sub_dict[uuid.hex] = int_id
            simple_dict[type_str] = sub_dict
        return simple_dict

    @classmethod
    def from_dict(cls, dict: Dict[str, Dict]) -> KeyIdMap:
        """
        Convert dict(bidict values to dictionary with dict values) into KeyIdMap.

        :return: KeyIdMap object
        :rtype: :class:`KeyIdMap`

        :Usage example:

         .. code-block:: python

            dict = {
                  "tags": {},
                  "objects": {},
                  "figures": {},
                  "videos": {}
              }
            key_id_map = KeyIdMap.from_dict(dict)
        """
        result = cls()
        for key_type, value_dict in dict.items():
            for key_str, id in value_dict.items():
                result._add(key_type, uuid.UUID(key_str), id)
        return result

    def dump_json(self, path: str) -> None:
        """
        Write KeyIdMap to file with given path.

        :param path: Target file path.
        :type path: str
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map.dump_json('/home/admin/work/projects/key_id.json')
        """
        simple_dict = self.to_dict()
        dump_json_file(simple_dict, path, indent=4)

    @classmethod
    def load_json(cls, path: str) -> KeyIdMap:
        """
        Decoding data from json file with given filename to KeyIdMap.

        :param path: Target file path.
        :type path: str
        :return: KeyIdMap object
        :rtype: :class:`KeyIdMap`

        :Usage example:

         .. code-block:: python

            new_key_id = KeyIdMap.load_json('/home/admin/work/projects/key_id.json')
        """
        simple_dict = load_json_file(path)
        result = cls()
        for key_type, value_dict in simple_dict.items():
            for key_str, id in value_dict.items():
                result._add(key_type, uuid.UUID(key_str), id)
        return result

    @classmethod
    def _add_to(cls, key_id_map, key_type, keys, ids):
        """
        Add given values(keys, ids) to KeyIdMap class object with given type of key
        :param key_id_map: KeyIdMap class object
        :param key_type: str
        :param keys: list of uuid class objects
        :param ids: list of integers
        :return: None if key_id_map parameter is None
        """
        if key_id_map is None:
            return
        for key, id in zip(keys, ids):
            key_id_map._add(key_type, key, id)

    @classmethod
    def add_tags_to(cls, key_id_map: KeyIdMap, keys: List[UUID], ids: List[int]) -> None:
        """
        Add :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` type of instances with given values(keys, ids) to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: List of UUID objects.
        :type key: List[UUID]
        :param id: List of :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` IDs.
        :type id: List[int]
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            uuid_1 = uuid.uuid4()
            tag_id_1 = 1213
            uuid_2 = uuid.uuid4()
            tag_id_2 = 3686
            KeyIdMap.add_tags_to(key_id_map, [uuid_1, uuid_2], [tag_id_1, tag_id_2])
        """
        cls._add_to(key_id_map, TAGS, keys, ids)

    @classmethod
    def add_tag_to(cls, key_id_map: KeyIdMap, key: UUID, id: int) -> None:
        """
        Add :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` type of instance with given key and id to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` ID.
        :type id: int
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4()
            new_tag_id = 1213
            KeyIdMap.add_tag_to(key_id_map, new_uuid, new_tag_id)
        """
        cls.add_tags_to(key_id_map, [key], [id])

    @classmethod
    def add_objects_to(cls, key_id_map: KeyIdMap, keys: List[UUID], ids: List[int]) -> None:
        """
        Add :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` type of instances with given values(keys, ids) to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: List of UUID objects.
        :type key: List[UUID]
        :param id: List of :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` IDs.
        :type id: List[int]
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            uuid_1 = uuid.uuid4()
            object_id_1 = 23
            uuid_2 = uuid.uuid4()
            object_id_2 = 57
            KeyIdMap.add_objects_to(key_id_map, [uuid_1, uuid_2], [object_id_1, object_id_2])
        """
        cls._add_to(key_id_map, OBJECTS, keys, ids)

    @classmethod
    def add_object_to(cls, key_id_map: KeyIdMap, key: UUID, id: int) -> None:
        """
        Add :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` type of instance with given key and id to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` ID.
        :type id: int
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4()
            new_object_id = 76
            KeyIdMap.add_object_to(key_id_map, new_uuid, new_object_id)
        """
        cls.add_objects_to(key_id_map, [key], [id])

    @classmethod
    def add_figures_to(cls, key_id_map: KeyIdMap, keys: List[UUID], ids: List[int]) -> None:
        """
        Add :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` type of instances with given values(keys, ids) to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: List of UUID objects.
        :type key: List[UUID]
        :param id: List of :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` IDs.
        :type id: List[int]
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            uuid_1 = uuid.uuid4()
            figure_id_1 = 23
            uuid_2 = uuid.uuid4()
            figure_id_2 = 57
            KeyIdMap.add_figures_to(key_id_map, [uuid_1, uuid_2], [figure_id_1, figure_id_2])
        """
        cls._add_to(key_id_map, FIGURES, keys, ids)

    @classmethod
    def add_figure_to(cls, key_id_map: KeyIdMap, key: UUID, id: int) -> None:
        """
        Add :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` type of instance with given key and id to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoFigure<supervisely.video_annotation.video_figure.VideoFigure>` ID.
        :type id: int
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4()
            new_figure_id = 3834
            KeyIdMap.add_figure_to(key_id_map, new_uuid, new_figure_id)
        """
        cls.add_figures_to(key_id_map, [key], [id])

    @classmethod
    def add_videos_to(cls, key_id_map: KeyIdMap, keys: List[UUID], ids: List[int]) -> None:
        """
        Add :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` type of instances with given values(keys, ids) to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: List of UUID objects.
        :type key: List[UUID]
        :param id: List of :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` IDs.
        :type id: List[int]
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            uuid_1 = uuid.uuid4()
            video_id_1 = 567
            uuid_2 = uuid.uuid4()
            video_id_2 = 5200
            KeyIdMap.add_videos_to(key_id_map, [uuid_1, uuid_2], [video_id_1, video_id_2])
        """
        cls._add_to(key_id_map, VIDEOS, keys, ids)

    @classmethod
    def add_video_to(cls, key_id_map: KeyIdMap, key: UUID, id: int) -> None:
        """
        Add :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` type of instance with given key and id to KeyIdMap object.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param key: UUID object.
        :type key: UUID
        :param id: :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>` ID.
        :type id: int
        :return: :class:`None`
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            key_id_map = KeyIdMap()
            new_uuid = uuid.uuid4()
            new_video_id = 3834
            KeyIdMap.add_video_to(key_id_map, new_uuid, new_video_id)
        """
        cls.add_videos_to(key_id_map, [key], [id])

    @classmethod
    def _merge_key_id_maps(
        cls, key_id_map_1: KeyIdMap, key_id_map_2: KeyIdMap, priority: int[1, 2] = 2
    ) -> KeyIdMap:
        """
        Merge two KeyIdMap objects and replace keys for identical IDs with keys of the prioritized KeyIdMap

        :param map_1: KeyIdMap object
        :type map_1: KeyIdMap
        :param map_2: KeyIdMap object
        :type map_2: KeyIdMap
        :param priority: From which KeyIdMap to save keys for identical IDs
        :type priority: int[1, 2]
        :return: Merged KeyIdMap
        :rtype: KeyIdMap
        :Usage example:

         .. code-block:: python

            key_id_map_1, key_id_map_2 = KeyIdMap(), KeyIdMap()
            new_uuid_1, new_uuid_2 = uuid.uuid4(), uuid.uuid4()
            new_video_id_1, new_video_id_2 = 3834, 4945

            KeyIdMap.add_video_to(key_id_map_1, new_uuid_1, new_video_id_1)
            KeyIdMap.add_video_to(key_id_map_2, new_uuid_2, new_video_id_2)

            merged_key_id_map = KeyIdMap.merge_key_id_maps(key_id_map_1, key_id_map_2)
        """
        inverted_dicts = []
        for map in [key_id_map_1, key_id_map_2]:
            map_dict = map.to_dict()
            inverted_dict = {}
            for key, nested_dict in map_dict.items():
                inverted_dict[key] = {value: subkey for subkey, value in nested_dict.items()}
            inverted_dicts.append(inverted_dict)

        if priority == 2:
            dict_1, dict_2 = inverted_dicts
        else:
            dict_2, dict_1 = inverted_dicts

        for key in dict_1:
            if key in dict_2 and isinstance(dict_1[key], Dict) and isinstance(dict_2[key], Dict):
                dict_1[key].update(dict_2[key])

        inverted_dict = {}
        for key, nested_dict in dict_1.items():
            inverted_dict[key] = {value: subkey for subkey, value in nested_dict.items()}

        return cls.from_dict(inverted_dict)



# coding: utf-8
from __future__ import annotations

import inspect
import io
import math
from functools import partial, wraps
from typing import Optional, Union

from tqdm import tqdm

from supervisely._utils import is_development, is_production, sizeof_fmt
from supervisely.sly_logger import EventType, logger


# float progress of training, since zero
def epoch_float(epoch, train_it, train_its):
    return epoch + train_it / float(train_its)


class Progress:
    """
    Modules operations monitoring and displaying statistics of data processing. :class:`Progress<Progress>` object is immutable.

    :param message: Progress message e.g. "Images uploaded:", "Processing:".
    :type message: str
    :param total_cnt: Total count.
    :type total_cnt: int, optional
    :param ext_logger: Logger object.
    :type ext_logger: logger, optional
    :param is_size: Shows Label size.
    :type is_size: bool, optional
    :param need_info_log: Shows info log.
    :type need_info_log: bool, optional
    :param min_report_percent: Minimum report percent of total items in progress to log.
    :type min_report_percent: int, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.sly_logger import logger

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        progress = sly.Progress("Images downloaded: ", len(img_infos), ext_logger=logger, is_size=True, need_info_log=True)
        api.image.download_paths(ds_id, image_ids, save_paths, progress_cb=progress.iters_done_report)

        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 0,
        #  "total": 6, "current_label": "0.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:45.659Z", "level": "info"}
        # {"message": "Images downloaded:  [0.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:45.660Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 1,
        #  "total": 6, "current_label": "1.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.134Z", "level": "info"}
        # {"message": "Images downloaded:  [1.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.134Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 2,
        #  "total": 6, "current_label": "2.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "Images downloaded:  [2.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 3,
        #  "total": 6, "current_label": "3.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "Images downloaded:  [3.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 4,
        #  "total": 6, "current_label": "4.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "Images downloaded:  [4.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.135Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 5,
        #  "total": 6, "current_label": "5.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.136Z", "level": "info"}
        # {"message": "Images downloaded:  [5.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.136Z", "level": "info"}
        # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Images downloaded: ", "current": 6,
        #  "total": 6, "current_label": "6.0 B", "total_label": "6.0 B", "timestamp": "2021-03-17T13:57:46.136Z", "level": "info"}
        # {"message": "Images downloaded:  [6.0 B / 6.0 B]", "timestamp": "2021-03-17T13:57:46.136Z", "level": "info"}
    """

    def __init__(
        self,
        message: str,
        total_cnt: Optional[int] = None,
        ext_logger: Optional[logger] = None,
        is_size: Optional[bool] = False,
        need_info_log: Optional[bool] = False,
        min_report_percent: Optional[int] = 1,
    ):
        self.is_size = is_size
        self.message = message
        self.total = total_cnt
        self.current = 0
        self.is_total_unknown = True if total_cnt in [None, 0] else False

        self.total_label = ""
        self.current_label = ""
        self._refresh_labels()

        self.reported_cnt = 0
        self.logger = logger if ext_logger is None else ext_logger
        self.report_every = max(1, math.ceil((total_cnt or 0) / 100 * min_report_percent))
        self.need_info_log = need_info_log

        mb5 = 5 * 1024 * 1024
        if self.is_size and self.is_total_unknown:
            self.report_every = mb5  # 5mb

        mb1 = 1 * 1024 * 1024
        if self.is_size and self.is_total_unknown is False and self.report_every < mb1:
            self.report_every = mb1  # 1mb

        if (
            self.is_size
            and self.is_total_unknown is False
            and self.total > 40 * 1024 * 1024
            and self.report_every < mb5
        ):
            self.report_every = mb5

        self.report_progress()

    def _refresh_labels(self):
        if self.is_size:
            self.total_label = (
                sizeof_fmt(self.current) if self.is_total_unknown else sizeof_fmt(self.total)
            )
            self.current_label = sizeof_fmt(self.current)
        else:
            self.total_label = str(self.current if self.is_total_unknown else self.total)
            self.current_label = str(self.current)

    def iter_done(self) -> None:
        """
        Increments the current iteration counter by 1
        """
        self.current += 1
        if self.is_total_unknown:
            self.total = self.current
        self._refresh_labels()

    def iters_done(self, count: int) -> None:
        """
        Increments the current iteration counter by given count

        :param count: Amount of iters
        :type count: int
        """
        self.current += count
        if self.is_total_unknown:
            self.total = self.current
        self._refresh_labels()

    def report_progress(self) -> None:
        """
        Logs a message with level INFO in logger. Message contain type of progress, subtask message, current and total number of iterations

        :return: None
        :rtype: :class:`NoneType`
        """
        self.print_progress()
        self.reported_cnt += 1

    def print_progress(self) -> None:
        """
        Logs a message with level INFO on logger. Message contain type of progress, subtask message, currtnt and total number of iterations
        """
        extra = {
            "event_type": EventType.PROGRESS,
            "subtask": self.message,
            "current": math.ceil(self.current),
            "total": math.ceil(self.current) if self.is_total_unknown else math.ceil(self.total),
        }

        if self.is_size:
            extra["current_label"] = self.current_label
            extra["total_label"] = self.total_label

        self.logger.info("progress", extra=extra)
        if self.need_info_log is True:
            self.logger.info(f"{self.message} [{self.current_label} / {self.total_label}]")

    def need_report(self) -> bool:
        if (
            (self.is_total_unknown)
            or (self.current >= self.total)
            or (self.current % self.report_every == 0)
            or ((self.reported_cnt - 1) < (self.current // self.report_every))
        ):
            return True
        return False

    def report_if_needed(self) -> None:
        """
        Determines whether the message should be logged depending on current number of iterations
        """
        if self.need_report():
            self.report_progress()

    def iter_done_report(self) -> None:  # finish & report
        """
        Increments the current iteration counter by 1 and logs a message depending on current number of iterations.

        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            progress = sly.Progress("Processing:", len(img_infos))
            for img_info in img_infos:
                img_names.append(img_info.name)
                progress.iter_done_report()

            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 0, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 1, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 2, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 3, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 4, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 5, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 6, "total": 6, "timestamp": "2021-03-17T14:29:33.207Z", "level": "info"}
        """
        self.iter_done()
        self.report_if_needed()

    def iters_done_report(self, count: int) -> None:  # finish & report
        """
        Increments the current iteration counter by given count and logs a message depending on current number of iterations.

        :param count: Counter.
        :type count: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            progress = sly.Progress("Processing:", len(img_infos))
            for img_info in img_infos:
                img_names.append(img_info.name)
                progress.iters_done_report(1)

            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 0, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 1, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 2, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 3, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 4, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 5, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Processing:",
            #  "current": 6, "total": 6, "timestamp": "2021-03-17T14:31:21.655Z", "level": "info"}
        """
        self.iters_done(count)
        self.report_if_needed()

    def set_current_value(self, value: int, report: Optional[bool] = True) -> None:
        """
        Increments the current iteration counter by this value minus the current value of the counter and logs a message depending on current number of iterations.

        :param value: Current value.
        :type value: int
        :param report: Defines whether to report to log or not.
        :type report: bool
        :return: None
        :rtype: :class:`NoneType`
        """
        if report is True:
            self.iters_done_report(value - self.current)
        else:
            self.iters_done(value - self.current)

    def set(self, current: int, total: Optional[int], report: Optional[bool] = True) -> None:
        """
        Sets counter current value and total value and logs a message depending on current number of iterations.

        :param current: Current count.
        :type current: int
        :param total: Total count.
        :type total: int, optional
        :param report: Defines whether to report to log or not.
        :type report: bool
        :return: None
        :rtype: :class:`NoneType`
        """
        self.total = total

        self.is_total_unknown = True if self.total in [None, 0] else False
        self.current = current
        self.reported_cnt = 0
        self.report_every = max(1, math.ceil((total or 0) / 100))
        self._refresh_labels()
        if report is True:
            self.report_if_needed()


def report_agent_rpc_ready() -> None:
    """
    Logs a message with level INFO on logger
    """
    logger.info("Ready to get events", extra={"event_type": EventType.TASK_DEPLOYED})


def report_import_finished() -> None:
    """
    Logs a message with level INFO on logger
    """
    logger.info("import finished", extra={"event_type": EventType.IMPORT_APPLIED})


def report_inference_finished() -> None:
    """
    Logs a message with level INFO on logger
    """
    logger.info("model applied", extra={"event_type": EventType.MODEL_APPLIED})


def report_dtl_finished() -> None:
    """
    Logs a message with level INFO on logger
    """
    logger.info("DTL finished", extra={"event_type": EventType.DTL_APPLIED})


def report_dtl_verification_finished(output: str) -> None:
    """
    Logs a message with level INFO on logger
    :param output: str
    """
    logger.info(
        "Verification finished.", extra={"output": output, "event_type": EventType.TASK_VERIFIED}
    )


def _report_metrics(m_type, epoch, metrics):
    logger.info(
        "metrics",
        extra={"event_type": EventType.METRICS, "type": m_type, "epoch": epoch, "metrics": metrics},
    )


def report_metrics_training(epoch, metrics):
    _report_metrics("train", epoch, metrics)


def report_metrics_validation(epoch, metrics):
    _report_metrics("val", epoch, metrics)


def report_checkpoint_saved(checkpoint_idx, subdir, sizeb, best_now, optional_data) -> None:
    logger.info(
        "checkpoint",
        extra={
            "event_type": EventType.CHECKPOINT,
            "id": checkpoint_idx,
            "subdir": subdir,
            "sizeb": sizeb,
            "best_now": best_now,
            "optional": optional_data,
        },
    )


class tqdm_sly(tqdm, Progress):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        # for self._upload_monitor()
        self._iteration_value = 0
        self._iteration_number = 0
        self._iteration_locked = False
        self._total_monitor_size = 0

        # self.n = None
        # self._total = kwargs.get("total", kwargs.get("total_cnt"))

        for _tqdm, _progress in {
            "total": "total_cnt",
            "desc": "message",
            "unit": "is_size",
            "unit_scale": "is_size",
        }.items():
            if kwargs.get(_tqdm) is not None and kwargs.get(_progress) is not None:
                raise ValueError(
                    f"Ambiguity error: Please specify only one of arguments: '{_tqdm}' or '{_progress}'."
                )

        kwargs_tqdm = self._handle_kwargs_tqdm(args, kwargs.copy())
        kwargs_tqdm.setdefault("unit_divisor", 1024)

        if is_development():
            tqdm.__init__(
                self,
                *args,
                **kwargs_tqdm,
            )
            self.offset = 0  # to prevent overfilling of tqdm in console
        else:
            for k, v in {
                "disable": False,  # do not disable for now
                "delay": 0,  # sec init delay
                "mininterval": 3,  # sec between reports
                "miniters": 0,
                "delay": 0,
            }.items():
                kwargs_tqdm.setdefault(k, v)

            tqdm.__init__(
                self,
                *args,
                **kwargs_tqdm,
            )
            self.disable = True  # now disable tqdm logging

            kwargs = self._handle_args_and_kwargs_prod(args, kwargs)
            Progress.__init__(
                self,
                **kwargs,
            )
            self.n = 0

    def __iter__(self):
        """Backward-compatibility to use: for x in tqdm(iterable)"""

        if is_development():
            yield from super().__iter__()
        else:
            # Inlining instance variables as locals (speed optimisation)
            iterable = self.iterable
            # self.disable = True

            mininterval = self.mininterval
            last_print_t = self.last_print_t
            last_print_n = self.last_print_n
            min_start_t = self.start_t + self.delay
            n = self.n
            time = self._time

            try:
                for obj in iterable:
                    yield obj
                    # Update and possibly print the progressbar.
                    # Note: does not call self.update(1) for speed optimisation.
                    n += 1
                    # self.iter_done_report()

                    if n - last_print_n >= self.miniters:
                        cur_t = time()
                        dt = cur_t - last_print_t
                        if dt >= mininterval and cur_t >= min_start_t:
                            Progress.need_report(self)
                            Progress.iters_done_report(self, (n - last_print_n))

                            last_print_n = self.n  # self.last_print_n
                            last_print_t = cur_t  # self.last_print_t
            finally:
                self.n = n
                self.close()

    def update(self, count):
        if is_development():
            if self.total is not None:
                count = min(count, self.total - self.offset)
                self.offset += count

            tqdm.update(self, count)

            if self.n == self.total:
                self.close()
        else:
            Progress.iters_done_report(self, count)
            self.n += count

    def __call__(
        self,
        *args,
        **kwargs,
    ):
        return self.update(
            *args,
            **kwargs,
        )

    def _progress_monitor(self, monitor):
        if is_development() and self.n >= self.total:
            self.refresh()
            self.close()

        if monitor.bytes_read == 8192:
            self._total_monitor_size += monitor.len

        if self._total_monitor_size > self.total:
            self.total = self._total_monitor_size

        if not self._iteration_locked:
            if is_development():
                super().update(self._iteration_value + monitor.bytes_read - self.n)
            else:
                self.set_current_value(self._iteration_value + monitor.bytes_read, report=False)

        if is_production() and self.need_report():
            self.report_progress()

        if monitor.bytes_read == monitor.len and not self._iteration_locked:
            self._iteration_value += monitor.len
            self._iteration_number += 1
            self._iteration_locked = True
            if is_development():
                self.refresh()

        if monitor.bytes_read < monitor.len:
            self._iteration_locked = False

    def get_partial(self):
        return partial(self._progress_monitor)

    def _handle_kwargs_tqdm(self, args, kwargs_tqdm):
        # pop and convert every possible (and relevant) kwarg from Progress
        if len(args) < 2:  # i.e. 'desc' not set as a positional argument
            if kwargs_tqdm.get("message") is not None:
                kwargs_tqdm.setdefault("desc", kwargs_tqdm["message"])
                kwargs_tqdm.pop("message")
            else:
                kwargs_tqdm.setdefault("desc", "Processing")
        if len(args) < 3:  # i.e. 'total' not set as a positional argument
            if kwargs_tqdm.get("total_cnt") is not None:
                kwargs_tqdm.setdefault("total", kwargs_tqdm["total_cnt"])
                kwargs_tqdm.pop("total_cnt")
        if len(args) < 12:  # i.e. 'unit' not set as a positional argument
            if kwargs_tqdm.pop("is_size", None) == True:
                kwargs_tqdm["unit"] = "B"
                kwargs_tqdm["unit_scale"] = True
        return kwargs_tqdm

    def _handle_args_and_kwargs_prod(self, args: tuple, kwargs: dict):
        # pop and convert every possible (and relevant) kwarg from tqdm
        # mention that tqdm is a prior parent class
        if len(args) < 2:  # i.e. 'desc' not set as a positional argument
            if kwargs.get("desc") is not None:
                kwargs.setdefault("message", kwargs["desc"])
                kwargs.pop("desc")
            else:
                kwargs.setdefault("message", "Processing")
        else:
            kwargs.setdefault("message", args[1])  # args[1]==desc
        if len(args) < 3:  # i.e. 'total' not set as a positional argument
            if kwargs.get("total") is not None:
                kwargs.setdefault("total_cnt", kwargs["total"])
                kwargs.pop("total")
        else:
            kwargs.setdefault("total_cnt", args[2])  # args[2]==total
        if len(args) < 12:  # i.e. 'unit' not set as a positional argument
            if kwargs.get("unit") in [
                "",
                "B",
                "k",
                "M",
                "G",
                "T",
                "P",
                "E",
                "Z",
            ] and kwargs.pop("unit_scale", None):
                kwargs["is_size"] = True
                kwargs.pop("unit")
        else:
            if (
                args[11] in ["", "B", "k", "M", "G", "T", "P", "E", "Z"] and args[12] == True
            ):  # i.e. unit=="B" and unit_scale==True
                kwargs["is_size"] = True

        tqdm_init_params = inspect.signature(tqdm.__init__).parameters.keys()
        for keyword in tqdm_init_params:
            if keyword in kwargs:
                kwargs.pop(keyword)

        # see original tqdm.__init__ for logic behaviour
        iterable_is_not_none = False if kwargs.get("iterable") is None and len(args) == 0 else True
        if kwargs.get("total_cnt") is None and iterable_is_not_none is True:
            try:
                iterable = kwargs.get("iterable", args[0])
                kwargs["total_cnt"] = len(iterable)
            except (TypeError, AttributeError):
                kwargs["total_cnt"] = None
        if kwargs.get("total_cnt") == float("inf"):
            # Infinite iterations, behave same as unknown
            kwargs["total_cnt"] = None

        return kwargs

    @classmethod
    def from_original_tqdm(
        cls,
        orig_tqdm: tqdm,
        **kwargs,
    ):
        # iterable=None, desc=None, total=None, leave=True, file=None,
        #          ncols=None, mininterval=0.1, maxinterval=10.0, miniters=None,
        #          ascii=None, disable=False, unit='it', unit_scale=False,
        #          dynamic_ncols=False, smoothing=0.3, bar_format=None, initial=0,
        #          position=None, postfix=None, unit_divisor=1000, write_bytes=False,
        #          lock_args=None, nrows=None, colour=None, delay=0, gui=False,

        vrs = vars(orig_tqdm).copy()
        sgn = inspect.signature(orig_tqdm.__init__)

        kw = {}

        non_idempotent_args = ["miniters", "position"]

        def _handle_ni_args(name, vrs):
            if name == "miniters":
                return None if vrs[name] == 0 else vrs[name]
            if name == "position":
                return None if vrs["pos"] == 0 else -vrs["pos"]

        for name, param in sgn.parameters.items():
            if name in kwargs:
                kw[name] = kwargs[name]
            elif name in non_idempotent_args:
                kw[name] = _handle_ni_args(name, vrs)
            else:
                if name == "kwargs":
                    pass
                else:
                    try:
                        kw[name] = vrs[name]
                    except KeyError:
                        kw[name] = param.default

        # legacy tqdm 'nested' in kwargs
        kw.update({k: v for k, v in kwargs.items() if k not in kw})
        return cls(**kw)


def handle_original_tqdm(func):
    # Deprecated
    @wraps(func)
    def wrapper_original_tqdm(*args, **kwargs):
        cb_name = (
            "progress_size_cb" if func.__qualname__ == "FileApi.upload_directory" else "progress_cb"
        )

        spc = inspect.getfullargspec(func)

        if cb_name not in spc.args:
            raise ValueError(
                f"The '{cb_name}' parameter was not found in the '{func.__qualname__}'"
            )
        else:  # Note: (args, kwargs) both in spc.args
            idx = spc.args.index(cb_name)
            try:
                progress_cb = args[idx]
            except IndexError:
                progress_cb = kwargs.get(cb_name)

            _progress_cb = progress_cb

            # Start progress bar setup
            if progress_cb is not None and isinstance(progress_cb, tqdm):
                if not type(progress_cb) == tqdm_sly:
                    progress_cb.clear()
                    _progress_cb = tqdm_sly.from_original_tqdm(progress_cb)

            new_args = list(args).copy()
            new_kwargs = kwargs.copy()
            try:
                new_args[idx] = _progress_cb
            except IndexError:
                new_kwargs[cb_name] = _progress_cb

        try:
            result = func(*new_args, **new_kwargs)
        except Exception as e:
            # Ensure progress bar gets closed in case of an exception
            if progress_cb is not None and isinstance(progress_cb, tqdm):
                if not type(progress_cb) == tqdm_sly:
                    progress_cb.close()
                _progress_cb.close()
            raise e

        # close progress bar
        if progress_cb is not None and isinstance(progress_cb, tqdm):
            if not type(progress_cb) == tqdm_sly:
                progress_cb.close()
            _progress_cb.close()

        return result

    return wrapper_original_tqdm



import functools
from fastapi import Request, BackgroundTasks
from typing import Any, Dict, Optional, Union
import supervisely as sly
from supervisely.nn.inference.tracking.tracker3d_interface import Tracker3DInterface
from supervisely.nn.inference import Inference
import os


class ObjectTracking3D(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        super().__init__(
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )
        self.load_on_device(model_dir, "cuda")

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    def serve(self):
        super().serve()
        server = self._app.get_server()

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error("An error occured:")
                    api.logger.exception(exc)

                    api.post(
                        "point-clouds.episodes.notify-annotation-tool",
                        data={
                            "type": "point-cloud-episodes:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request = None):
            # initialize tracker 3d interface
            context = request.state.context
            api: sly.Api = request.state.api
            self.pcd_interface = Tracker3DInterface(
                context=context,
                api=api,
            )
            api.logger.info("Starting tracking process")
            # propagate frame-by-frame
            for geom_idx in range(len(self.pcd_interface.geometries)):
                for i, pc_id in enumerate(self.pcd_interface.pc_ids):
                    # download input data
                    frame = {}
                    cloud_info = api.pointcloud.get_info_by_id(pc_id)
                    pcd_path = os.path.join(self.pcd_interface.pc_dir, cloud_info.name)
                    if geom_idx == 0:
                        api.pointcloud_episode.download_path(cloud_info.id, pcd_path)
                        self.pcd_interface._notify(task="load frame")
                    frame["pcd"] = pcd_path
                    if i == 0:
                        geometry = self.pcd_interface.geometries[geom_idx]
                        frame["bbox"] = geometry
                    # pass input data to model and get prediction
                    if i == len(self.pcd_interface.pc_ids) - 1:
                        predicted_cuboid = self.predict(frame, is_last_frame=True)
                    else:
                        predicted_cuboid = self.predict(frame)
                    # add predicted cuboid to frame
                    if i != 0:
                        pcd_id = self.pcd_interface.pc_ids[i]
                        fig_id = self.pcd_interface.figure_ids[geom_idx]
                        obj_id = api.pointcloud.figure.get_info_by_id(fig_id).object_id
                        track_id = self.pcd_interface.track_id
                        self.pcd_interface.add_cuboid_on_frame(
                            pcd_id, obj_id, predicted_cuboid.to_json(), track_id
                        )
            # clean directory with downloaded pointclouds
            sly.fs.clean_dir(self.pcd_interface.pc_dir)
            api.logger.info("Successfully finished tracking process")



import numpy as np
from typing import Generator, Optional, List, Callable, Dict
from typing import OrderedDict as OrderedDictType
from collections import OrderedDict

import supervisely as sly
from supervisely.geometry.geometry import Geometry
from logging import Logger


class TrackerInterface:
    def __init__(
        self,
        context,
        api,
        load_all_frames=False,
        notify_in_predict=False,
        per_point_polygon_tracking=True,
        frame_loader: Callable[[sly.Api, int, int], np.ndarray] = None,
    ):
        self.api: sly.Api = api
        self.logger: Logger = api.logger
        self.frame_index = context["frameIndex"]
        self.frames_count = context["frames"]

        self.track_id = context["trackId"]
        self.video_id = context["videoId"]
        self.object_ids = list(context["objectIds"])
        self.figure_ids = list(context["figureIds"])
        self.direction = context["direction"]

        # all geometries
        self.stop = len(self.figure_ids) * self.frames_count
        self.global_pos = 0
        self.global_stop_indicatior = False

        self.geometries: OrderedDictType[int, Geometry] = OrderedDict()
        self.frames_indexes: List[int] = []
        self._cur_frames_indexes: List[int] = []
        self._frames: Optional[np.ndarray] = None
        self.load_all_frames = load_all_frames
        self.per_point_polygon_tracking = per_point_polygon_tracking

        # increase self.stop by num of frames will be loaded
        self._add_frames_indexes()

        # increase self.stop by num of points
        self._add_geometries()

        self._hot_cache: Dict[int, np.ndarray] = {}
        self._local_cache_loader = frame_loader

        if self.load_all_frames:
            if notify_in_predict:
                self.stop += self.frames_count + 1
            self._load_frames()

    def add_object_geometries(self, geometries: List[Geometry], object_id: int, start_fig: int):
        for frame, geometry in zip(self._cur_frames_indexes[1:], geometries):
            if self.global_stop_indicatior:
                self._notify(True, task="stop tracking")
                break
            self.add_object_geometry_on_frame(geometry, object_id, frame)

        self.geometries[start_fig] = geometries[-1]

    def frames_loader_generator(self) -> Generator[None, None, None]:
        if self.load_all_frames:
            self._cur_frames_indexes = self.frames_indexes
            yield
            return

        ind = self.frames_indexes[0]
        frame = self._load_frame(ind)
        for next_ind in self.frames_indexes[1:]:
            next_frame = self._load_frame(next_ind)
            self._frames = np.array([frame, next_frame])
            self.frames_count = 1
            self._cur_frames_indexes = [ind, next_ind]
            yield
            frame = next_frame
            ind = next_ind

            if self.global_stop_indicatior:
                self.clear_cache()
                return

    def add_object_geometry_on_frame(
        self,
        geometry: Geometry,
        object_id: int,
        frame_ind: int,
        notify: bool = True,
    ):
        self.api.video.figure.create(
            self.video_id,
            object_id,
            frame_ind,
            geometry.to_json(),
            geometry.geometry_name(),
            self.track_id,
        )
        self.logger.debug(f"Added {geometry.geometry_name()} to frame #{frame_ind}")
        if notify:
            self._notify(task="add geometry on frame")

    def clear_cache(self):
        self._hot_cache.clear()

    def _add_geometries(self):
        self.logger.info("Adding geometries.")
        points = 0
        for figure_id in self.figure_ids:
            figure = self.api.video.figure.get_info_by_id(figure_id)
            geometry = sly.deserialize_geometry(figure.geometry_type, figure.geometry)
            self.geometries[figure_id] = geometry

            self.api.logger.debug(f"Added {figure.geometry_type} #{figure_id}")

            # per point track notification
            if isinstance(geometry, sly.Point):
                points += 1
            elif isinstance(geometry, sly.Polygon):
                points += len(geometry.exterior) + len(geometry.interior)
            elif isinstance(geometry, sly.GraphNodes):
                points += len(geometry.nodes.items())
            elif isinstance(geometry, sly.Polyline):
                points += len(geometry.exterior)

        if self.per_point_polygon_tracking:
            if not self.load_all_frames:
                self.stop += points * self.frames_count
            else:
                self.stop += points

        self.logger.info("Geometries added.")
        # TODO: other geometries

    def _add_frames_indexes(self):
        total_frames = self.api.video.get_info_by_id(self.video_id).frames_count
        cur_index = self.frame_index

        while 0 <= cur_index < total_frames and len(self.frames_indexes) < self.frames_count + 1:
            self.frames_indexes.append(cur_index)
            cur_index += 1 if self.direction == "forward" else -1

        if self.load_all_frames:
            self.stop += len(self.frames_indexes)

    def _load_frame(self, frame_index):
        if frame_index in self._hot_cache:
            return self._hot_cache[frame_index]
        if self._local_cache_loader is None:
            self._hot_cache[frame_index] = self.api.video.frame.download_np(
                self.video_id, frame_index
            )
        else:
            self._hot_cache[frame_index] = self._local_cache_loader(
                self.api, self.video_id, frame_index
            )
        return self._hot_cache[frame_index]

    def _load_frames(self):
        rgbs = []
        self.logger.info(f"Loading {len(self.frames_indexes)} frames.")

        for frame_index in self.frames_indexes:
            img_rgb = self._load_frame(frame_index)
            rgbs.append(img_rgb)
            self._notify(task="load frame")
        self._frames = rgbs
        self.logger.info("Frames loaded.")

    def _notify(
        self,
        stop: bool = False,
        fstart: Optional[int] = None,
        fend: Optional[int] = None,
        task: str = "not defined",
    ):
        self.global_pos += 1

        if stop:
            pos = self.stop
        else:
            pos = self.global_pos

        fstart = min(self.frames_indexes) if fstart is None else fstart
        fend = max(self.frames_indexes) if fend is None else fend

        self.logger.debug(f"Task: {task}")
        self.logger.debug(f"Notification status: {pos}/{self.stop}")

        self.global_stop_indicatior = self.api.video.notify_progress(
            self.track_id,
            self.video_id,
            fstart,
            fend,
            pos,
            self.stop,
        )

        self.logger.debug(f"Notification status: stop={self.global_stop_indicatior}")

        if self.global_stop_indicatior and self.global_pos < self.stop:
            self.logger.info("Task stoped by user.")

    @property
    def frames(self) -> np.ndarray:
        return self._frames

    @property
    def frames_with_notification(self) -> np.ndarray:
        """Use this in prediction."""
        self._notify(task="get frames")
        return self._frames



import unittest
from unittest.mock import patch

from supervisely.api.api import Api
from supervisely.api.project_api import ProjectType
from supervisely.video.video import UnsupportedVideoFormat


class TestVideoApi(unittest.TestCase):
    workspace_id = None
    project_id = None
    api = None

    @classmethod
    def setUpClass(cls):
        cls.api = Api.from_env()
        workspace_id = input("Enter workspace ID for tests here >> ")
        with patch("builtins.input", return_value="user_input_value"):
            cls.workspace_id = int(workspace_id)

    def setUp(self):
        self.project_api = self.api.project
        self.dataset_api = self.api.dataset
        self.video_api = self.api.video
        self.valid_name = "valid.mp4"
        self.invalid_name_wo_ext = "invalid"
        self.invalid_name_ext = "invalid.avi"
        # replace with paths for existing files
        self.path_valid_ext = "/video.mp4"
        self.path_invalid_ext = "/video.png"

    def tearDown(self):
        self.api.project.remove(self.project_id)

    def create_test_projects(self, count):
        """
        Create test projects for testing.
        """
        created_project_ids = []
        for i in range(count):
            created_project = self.project_api.create(
                self.workspace_id,
                name=f"[UT] Video Project {i+1}",
                type=ProjectType.VIDEOS,
                change_name_if_conflict=True,
            )
            created_project_ids.append(created_project.id)
        return created_project_ids

    def create_test_datasets(self, count):
        """
        Create test datasets for testing.
        """
        created_dataset_ids = []
        for i in range(count):
            created_dataset = self.dataset_api.create(
                self.project_id,
                name=f"[UT] Dataset {i+1}",
                change_name_if_conflict=True,
            )
            created_dataset_ids.append(created_dataset.id)
        return created_dataset_ids

    def test_upload_paths(self):
        self.project_id = self.create_test_projects(1)[0]
        created_datase_id = self.create_test_datasets(1)[0]
        # Verify if extension is supported (file path)
        with self.assertRaises(UnsupportedVideoFormat):
            self.video_api.upload_paths(
                created_datase_id, [self.invalid_name_wo_ext], [self.path_invalid_ext]
            )
        # Verify if extensions in the file name and path match
        with self.assertRaises(ValueError):
            self.video_api.upload_paths(
                created_datase_id, [self.invalid_name_ext], [self.path_valid_ext]
            )
        self.assertIsNotNone(
            self.video_api.upload_paths(created_datase_id, [self.valid_name], [self.path_valid_ext])
        )

    def test_upload_links_with_valid_links(self):
        self.project_id = self.create_test_projects(1)[0]
        created_datase_id = self.create_test_datasets(1)[0]
        video_info = self.video_api.upload_paths(
            created_datase_id, [self.valid_name], [self.path_valid_ext]
        )[0]
        hash = video_info.hash
        name = "new_" + self.valid_name
        link_video_info = self.video_api.upload_hashes(created_datase_id, [name], [hash])[0]
        self.assertIsNotNone(link_video_info)
        with self.assertRaises(UnsupportedVideoFormat):
            self.video_api.upload_hashes(created_datase_id, [self.invalid_name_wo_ext], [hash])


if __name__ == "__main__":
    unittest.main()



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from typing import Callable, Optional

from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class Button(Widget):
    """Button widget in Supervisely is a user interface element that allows users to create clickable buttons in the applications.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/controls/button>`_
        (including screenshots and examples).

    :param text: Text to be displayed on the button.
    :type text: Optional[str]
    :param button_type: Type of the button.
    :type button_type: Optional[Literal["primary", "info", "warning", "danger", "success", "text"]]
    :param button_size: Size of the button.
    :type button_size: Optional[Literal["mini", "small", "large"]]
    :param plain: If True, the button will be plain.
    :type plain: Optional[bool]
    :param show_loading: If True, the button will show loading animation.
    :type show_loading: Optional[bool]
    :param icon: Icon to be displayed on the button. E.g. "zmdi zmdi-play", "zmdi zmdi-stop", "zmdi zmdi-pause".
        List of available icons can be found `here <http://zavoloklom.github.io/material-design-iconic-font/icons.html>`_.
    :type icon: Optional[str]
    :param icon_gap: Gap between the icon and the text in pixels.
    :type icon_gap: Optional[int]
    :param widget_id: Unique widget identifier.
    :type widget_id: Optional[str]
    :param link: Link to be opened on button click.
    :type link: Optional[str]
    :param emit_on_click: Name of the event to be emitted on button click.
    :type emit_on_click: Optional[str]
    :param style: CSS style to be applied to the button.
    :type style: Optional[str]
    :param call_on_click: Function to be called on button click.
    :type call_on_click: Optional[str]

    :Usage example:
    .. code-block:: python
        from supervisely.app.widgets import Button

        button = Button(text="Button", button_type="primary", button_size="large")
    """

    class Routes:
        CLICK = "button_clicked_cb"

    def __init__(
        self,
        text: Optional[str] = "Button",
        button_type: Optional[
            Literal["primary", "info", "warning", "danger", "success", "text"]
        ] = "primary",
        button_size: Optional[Literal["mini", "small", "large"]] = None,
        plain: Optional[bool] = False,
        show_loading: Optional[bool] = True,
        icon: Optional[str] = None,
        icon_gap: Optional[int] = 5,
        widget_id: Optional[str] = None,
        link: Optional[str] = None,
        emit_on_click: Optional[str] = None,
        style: Optional[str] = None,
        call_on_click: Optional[str] = None,
    ):
        self._widget_routes = {}

        self._text = text
        self._button_type = button_type
        self._button_size = button_size
        self._plain = plain
        self._icon_gap = icon_gap
        self._link = link
        if icon is None:
            self._icon = ""
        else:
            self._icon = f'<i class="{icon}" style="margin-right: {icon_gap}px"></i>'

        self._loading = False
        self._disabled = False
        self._show_loading = show_loading
        self._click_handled = False
        self._emit_on_click = emit_on_click
        self._style = style
        self._call_on_click = call_on_click

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - text: Text to be displayed on the button.
            - button_type: Type of the button.
            - plain: If True, the button will be plain.
            - button_size: Size of the button.
            - loading: If True, the button will show loading animation.
            - disabled: If True, the button will be disabled.
            - icon: Icon to be displayed on the button.
            - link: Link to be opened on button click.
        """
        return {
            "text": self._text,
            "button_type": self._button_type,
            "plain": self._plain,
            "button_size": self._button_size,
            "loading": self._loading,
            "disabled": self._disabled,
            "icon": self._icon,
            "link": self._link,
        }

    def get_json_state(self) -> None:
        """Button widget doesn't have state, so this method returns None."""
        return None

    @property
    def text(self) -> str:
        """Returns the text to be displayed on the button.

        :return: Text to be displayed on the button.
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """Sets the text to be displayed on the button.

        :param value: Text to be displayed on the button.
        :type value: str
        """
        self._text = value
        DataJson()[self.widget_id]["text"] = self._text
        DataJson().send_changes()

    @property
    def icon(self) -> str:
        """Returns the icon to be displayed on the button.

        :return: Icon to be displayed on the button.
        :rtype: str
        """
        return self._icon

    @icon.setter
    def icon(self, value: str) -> None:
        """Sets the icon to be displayed on the button.

        :param value: Icon to be displayed on the button.
        :type value: str
        """
        if value is None:
            self._icon = ""
        else:
            self._icon = f'<i class="{value}" style="margin-right: {self._icon_gap}px"></i>'
        DataJson()[self.widget_id]["icon"] = self._icon
        DataJson().send_changes()

    @property
    def button_type(self) -> str:
        """Returns the type of the button.

        :return: Type of the button.
        :rtype: str
        """
        return self._button_type

    @button_type.setter
    def button_type(
        self, value: Literal["primary", "info", "warning", "danger", "success", "text"]
    ) -> None:
        """Sets the type of the button.

        :param value: Type of the button.
        :type value: Literal["primary", "info", "warning", "danger", "success", "text"]
        """
        self._button_type = value
        DataJson()[self.widget_id]["button_type"] = self._button_type
        DataJson().send_changes()

    @property
    def plain(self) -> bool:
        """Returns True if the button is plain, False otherwise.

        :return: True if the button is plain, False otherwise.
        :rtype: bool
        """
        return self._plain

    @plain.setter
    def plain(self, value: bool) -> None:
        """Sets the button to be plain or not.

        :param value: If True, the button will be plain.
        :type value: bool
        """
        self._plain = value
        DataJson()[self.widget_id]["plain"] = self._plain
        DataJson().send_changes()

    @property
    def link(self) -> str:
        """Returns the link to be opened on button click.

        :return: Link to be opened on button click.
        :rtype: str
        """
        return self._link

    @link.setter
    def link(self, value: str) -> None:
        """Sets the link to be opened on button click.

        :param value: Link to be opened on button click.
        :type value: str
        """
        self._link = value
        DataJson()[self.widget_id]["link"] = self._link
        DataJson().send_changes()

    @property
    def loading(self) -> bool:
        """Returns True if the button shows loading animation, False otherwise.

        :return: True if the button shows loading animation, False otherwise.
        :rtype: bool
        """
        return self._loading

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets the button loading animation.

        :param value: If True, the animation will be enabled, otherwise disabled.
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    @property
    def show_loading(self) -> bool:
        """Returns True if the button shows loading animation, False otherwise.

        :return: True if the button shows loading animation, False otherwise.
        :rtype: bool
        """
        return self._show_loading

    @property
    def disabled(self) -> bool:
        """Returns True if the button is disabled, False otherwise.

        :return: True if the button is disabled, False otherwise.
        :rtype: bool
        """
        return self._disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        """Sets the button to be disabled or not.

        :param value: If True, the button will be disabled.
        :type value: bool
        """
        self._disabled = value
        DataJson()[self.widget_id]["disabled"] = self._disabled

    def click(self, func: Callable[[], None]) -> Callable[[], None]:
        """Decorator that allows to handle button click. Decorated function
        will be called on button click.

        :param func: Function to be called on button click.
        :type func: Callable
        :return: Decorated function.
        :rtype: Callable
        """
        route_path = self.get_route_path(Button.Routes.CLICK)
        server = self._sly_app.get_server()
        self._click_handled = True

        @server.post(route_path)
        def _click():
            if self.show_loading:
                self.loading = True
            try:
                func()
            except Exception as e:
                if self.show_loading and self.loading:
                    self.loading = False
                raise e
            if self.show_loading:
                self.loading = False

        return _click

    def _get_on_click(self):
        on_click_actions = []
        if self._emit_on_click:
            on_click_actions.append(f"$emit('{self._emit_on_click}');")

        if self._call_on_click:
            on_click_actions.append(f"{self._call_on_click}")

        if self._click_handled:
            on_click_actions.append(f"post('/{self.widget_id}/button_clicked_cb');")
        return " ".join(on_click_actions) if on_click_actions else None
